# SPDX-License-Identifier: GNU GPL v3
"""
Pyside6 (GUI components) main controller class.
"""

import logging
import numpy as np
from PySide6.QtWidgets import QApplication
from PySide6.QtCore import Signal, Slot, QObject
from sgtlib.modules import ProgressData, TaskResult

from .theme_manager import ThemeManager
from .network_controller import NetworkController
from ..workers.base_workers import BaseWorker
from ..workers.persistent_worker import PersistentProcessWorker
from ... import __version__, __title__
from ...compute.response_analyzer import ResponseAnalyzer


class MainController(QObject):

    _waitChanged = Signal()
    _waitTextChanged = Signal()
    errorSignal = Signal(str)
    showAlertSignal = Signal(str, str)
    syncModelSignal = Signal(object)
    updateProgressSignal = Signal(int, str)
    taskTerminatedSignal = Signal(bool, list)

    def __init__(self, qml_app: QApplication, parent: QObject = None):
        super().__init__(parent)
        self._qml_app = qml_app

        # Initialize flags
        self._wait_flag = False
        self._wait_msg = ""

        # Create network objects
        self._config_file = ""
        self._rgt_obj = ResponseAnalyzer(self._config_file)

        # Add Controllers
        self.theme_ctrl = ThemeManager()
        self.network_ctrl = NetworkController(self)

        # Create Persistent Workers (Processes)
        self._rgt_worker = PersistentProcessWorker(worker_id=1)

    @property
    def wait_flag(self) -> bool:
        """Returns the wait flag indicating if the application is currently running a task in the background."""
        return self._wait_flag

    @wait_flag.setter
    def wait_flag(self, value: bool):
        """Sets the wait flag indicating if the application is currently running a task in the background."""
        self._wait_flag = value

    @property
    def wait_msg(self) -> str:
        """Returns the wait message indicating the current task."""
        return self._wait_msg

    @wait_msg.setter
    def wait_msg(self, value: str):
        """Sets the wait message indicating the current task."""
        self._wait_msg = value

    @property
    def rgt_obj(self):
        return self._rgt_obj

    def _cancel_loading(self, worker_id):
        if worker_id == 1:
            self.network_ctrl.stop_task()

    def cleanup_workers(self):
        """Stop all persistent workers before app exit."""
        self.showAlertSignal.emit("Important Alert", "Please wait as we safely close the app...")
        for worker in [self._rgt_worker]:
            if worker:
                worker.stop()

    def handle_progress_update(self, status_data: ProgressData) -> None:
        """
        Handler function for progress updates for ongoing GT tasks.
        Args:
            status_data: ProgressData object that contains the percentage and status message of the current task.

        Returns:

        """

        if status_data is None:
            return

        if 0 <= status_data.percent <= 100:
            self.updateProgressSignal.emit(status_data.percent, status_data.message)
            logging.info(f"({status_data.sender}) {status_data.percent}%: {status_data.message}", extra={'user': 'RGT Logs'})

        if status_data.type == "info":
            self.updateProgressSignal.emit(101, status_data.message)
            logging.info(f"({status_data.sender}) {status_data.message}", extra={'user': 'RGT Logs'})
        elif status_data.type == "error":
            self.errorSignal.emit(status_data.message)
            logging.exception(f"({status_data.sender}) {status_data.message}", extra={'user': 'RGT Logs'})

    def handle_finished(self, worker_id: int, success_val: bool, result: None | TaskResult | list) -> None:
        """
        Handler function for sending updates/signals on termination of tasks.
        Args:
            worker_id: The process worker ID.
            success_val: True if the task was successful, False otherwise.
            result: The result of the task.
        Returns:
            None
        """
        self._cancel_loading(worker_id)
        if not success_val:
            if type(result) is list:
                logging.info(f"{result[0]} : {result[1]}", extra={'user': 'RGT Logs'})
                self.taskTerminatedSignal.emit(success_val, result)
        else:
            if isinstance(result, TaskResult):
                self.stop_current_task(cancel_job=False)
                if result.task_id == "Upload CSV":
                    upload_type = result.data[0]
                    file_path = result.data[1]
                    graph_data = result.data[2]

                    if type(file_path) is str:
                        self.rgt_obj._save_path = file_path

                    if upload_type == "vertices":
                        node_positions = graph_data
                        if node_positions is not None:
                            # flips vertically to have the same orientation as the initial image
                            y_coords, x_coords = zip(*node_positions)
                            neg_y_coords = [y * -1 for y in y_coords]
                            self.rgt_obj.list_data["vertex_positions"]["data"] = node_positions
                            self.rgt_obj.list_data["vertex_positions"]["value"] = 1
                            self.rgt_obj.list_data["flipped_vertex_coordinates"]["data"] = np.array(list(zip(x_coords, neg_y_coords)))
                            self.rgt_obj.list_data["flipped_vertex_coordinates"]["value"] = 1
                    elif upload_type == "edges":
                        edges = graph_data
                        if edges is not None:
                            self.rgt_obj.edge_list = edges
                            self.rgt_obj.list_data["edge_list"]["data"] = edges
                            self.rgt_obj.list_data["edge_list"]["value"] = 1
                    elif upload_type == "Potential List":
                        imposed_vertices = graph_data
                        if imposed_vertices is not None:
                            self.rgt_obj.list_data["given_potential_list"]["data"] = imposed_vertices
                            self.rgt_obj.list_data["given_potential_list"]["type"] = "File"
                            self.rgt_obj.list_data["given_potential_list"]["value"] = 1
                    elif upload_type == "Vertex List":
                        vert_list = graph_data
                        if vert_list is not None:
                            self.rgt_obj.list_data["vertex_list"]["data"] = vert_list
                            self.rgt_obj.list_data["vertex_list"]["type"] = "File"
                            self.rgt_obj.list_data["vertex_list"]["value"] = 1
                    elif upload_type == "Resistivity List":
                        res_list = graph_data
                        if res_list is not None:
                            self.rgt_obj.list_data["resistivity_list"]["data"] = res_list
                            self.rgt_obj.list_data["resistivity_list"]["value"] = 1
                    elif upload_type == "Inductance List":
                        ind_list = graph_data
                        if ind_list is not None:
                            self.rgt_obj.list_data["inductance_list"]["data"] = ind_list
                            self.rgt_obj.list_data["inductance_list"]["value"] = 1
                    elif upload_type == "Capacitance List":
                        cap_list = graph_data
                        if cap_list is not None:
                            self.rgt_obj.list_data["capacitance_list"]["data"] = cap_list
                            self.rgt_obj.list_data["capacitance_list"]["value"] = 1
                    elif upload_type == "Leak Resistivity List":
                        leak_list = graph_data
                        if leak_list is not None:
                            self.rgt_obj.list_data["leak_resistivity_list"]["data"] = leak_list
                            self.rgt_obj.list_data["leak_resistivity_list"]["value"] = 1
                    self.handle_progress_update(ProgressData(percent=100, sender="RGT", message=f"File Uploaded!"))
                    self.syncModelSignal.emit(self.rgt_obj)  # Sync models and refresh image
                    self.network_ctrl.imageChangedSignal.emit()  # trigger QML UI update
                    self.taskTerminatedSignal.emit(success_val, [])  # Hide Alert-Dialog
                if result.task_id == "Save Results":
                    # Saving files to Output Folder
                    self.handle_progress_update(ProgressData(percent=100, sender="RGT", message=f"Files Saved!"))
                    self.taskTerminatedSignal.emit(success_val, ["Files Saved", result.message])
                if result.task_id == "Compute Response":
                    new_rgt_obj = result.data
                    self.rgt_obj.copy_rgt_obj(new_rgt_obj)
                    self.handle_progress_update(ProgressData(percent=100, sender="RGT", message=result.message))

                    self.syncModelSignal.emit(self.rgt_obj) # Sync models and refresh image
                    self.network_ctrl.changeImageSignal.emit() # trigger QML UI update (load image)
                    self.taskTerminatedSignal.emit(success_val, [])  # Hide Alert-Dialog
            else:
                self.taskTerminatedSignal.emit(success_val, [])

    def submit_job(self, worker_id, task_fxn, fxn_args=(), track_updates: bool = True) -> None:
        """Start a background thread and its associated worker."""

        def _sync_signals(bg_worker: PersistentProcessWorker):
            bg_worker.taskCompleted.connect(self.handle_finished)
            if track_updates:
                bg_worker.inProgress.connect(self.handle_progress_update)

        if task_fxn is None or worker_id is None:
            return

        base_funcs = BaseWorker()

        if task_fxn == "Compute Response":
            target = base_funcs.task_compute_response
        elif task_fxn == "Upload CSV":
            target = base_funcs.task_upload_csv
        elif task_fxn == "Save Results":
            target = base_funcs.task_save_results
        else:
            return

        if worker_id == 1:
            started = self._rgt_worker.submit_task(func=target, args=fxn_args)
            if not started:
                self.showAlertSignal.emit("Please Wait", "Another GT job is running!")
                return
            _sync_signals(self._rgt_worker)
        else:
            return

    @Slot(int)
    def stop_current_task(self, worker_id: int = 1, cancel_job: bool = True):
        """Stop a background thread and its associated worker."""
        # self.showAlertSignal.emit("Important Alert", "Cancelling job, please wait...")
        if worker_id == 1:
            if cancel_job:
                self.handle_progress_update(
                    ProgressData(percent=99, sender="RGT", message="Cancelling job, please wait..."))
            else:
                # Restart Process after 3 tasks
                if self._rgt_worker.task_count < 3:
                    return
            self._rgt_worker.stop()
            self._rgt_worker = PersistentProcessWorker(worker_id)
            self.handle_finished(worker_id, True, None)

    @Slot(result=str)
    def get_about_details(self):
        about_str = f"""{__title__} v{__version__}"""
        self.network_ctrl.imageChangedSignal.emit()
        return about_str

    @Slot(result=str)
    def get_app_title(self):
        return f"{__title__}"

    @Slot(result=str)
    def get_app_version(self):
        """"""
        return f"v{__version__}"

    @Slot(result=bool)
    def is_task_running(self):
        return self._wait_flag

    @Slot()
    def reset_rgt_obj(self):
        self.rgt_obj.reset()
        self.network_ctrl._graph_loaded = False
        self.network_ctrl.imageChangedSignal.emit()

    @Slot()
    def load_graph_into_view(self):
        """Load the graph into the view."""
        try:
            self.network_ctrl.changeImageSignal.emit()
        except Exception as err:
            # self.reset_rgt_obj()
            self.network_ctrl.imageChangedSignal.emit()
            logging.exception("View Error: %s", err, extra={'user': 'RGT Logs'})
            self.showAlertSignal.emit("Graph Error", "Error loading graph. Try again.")



# SPDX-License-Identifier: GNU GPL v3
"""
Pyside6 (GUI components) controller class for network retrieval and computations.
"""

import ast
import logging
import numpy as np
from sgtlib.modules import ProgressData
from PySide6.QtCore import Signal, Slot, QObject

from ..models.checkbox_model import CheckBoxModel
from ...utils.config_loader import get_metric_options
from ...compute.response_analyzer import ALLOWED_GRAPH_FILE_EXTENSIONS


class NetworkController(QObject):


    changeImageSignal = Signal()
    imageChangedSignal = Signal()

    def __init__(self, controller_obj, parent: QObject = None):
        super().__init__(parent)
        self._ctrl = controller_obj
        self._graph_loaded = False
        self._applying_changes = False

        # Create Models
        self.rgtResponseOptions = CheckBoxModel([])

        # Electrical Response
        self.rgtFiles = CheckBoxModel([])
        self.rgtParamOptions = CheckBoxModel([])
        self.rgtDCParams = CheckBoxModel([])
        self.rgtPotentialOptions = CheckBoxModel([])
        self.rgtPotentialDirections = CheckBoxModel([])
        self.metricsModel = CheckBoxModel(get_metric_options())

        # Mechanical Response
        self.rgtPinnedDirections = CheckBoxModel([])
        self.rgtDisplacedDirections = CheckBoxModel([])

        # Attach listener for syncing models
        self._ctrl.syncModelSignal.connect(self.synchronize_rgt_models)

    def start_task(self, msg: str = "please wait..."):
        """Activate the wait flag and send a wait signal."""
        self._ctrl.wait_msg = msg
        self._ctrl.wait_flag = True

    def stop_task(self):
        """Deactivate the wait flag and send a wait signal."""
        self._ctrl.wait_msg = ""
        self._ctrl.wait_flag = False

    def synchronize_rgt_models(self, rgt_obj):
        """
        Reload ResponseGT configuration selections and controls from saved dict to QML components at every sync signal.

        :param rgt_obj: ResponseGT object with all the saved user-selected configuration selections.
        """
        if rgt_obj is None:
            return

        try:
            # Models Auto-update with saved sgt_obj configs. No need to re-assign!
            options_rgt = rgt_obj.configs
            list_data = rgt_obj.list_data

            # Get data from object configs
            rgt_response_type = [v for v in options_rgt.values() if v["type"] == "rgt-settings"]

            # (electrical response)
            rgt_files = [v for v in list_data.values() if (v["visible"] == 1 and v["value"] != 0)]
            rgt_param_type = [v for v in options_rgt.values() if v["type"] == "param-settings"]
            rgt_dc_params = [v for v in options_rgt.values() if v["type"] == "dc-param"]
            rgt_potentials = [v for v in options_rgt.values() if v["type"] == "potential-settings"]
            rgt_pot_dir = options_rgt["potential_direction"]["items"]

            # (mechanical response)
            rgt_pin_dir = options_rgt["pinned_direction"]["items"]
            rgt_disp_dir = options_rgt["displacement_direction"]["items"]

            # Update QML adapter-models with fetched data
            self.rgtResponseOptions.reset_data(rgt_response_type)
            self.rgtFiles.reset_data(rgt_files)
            self.rgtParamOptions.reset_data(rgt_param_type)
            self.rgtDCParams.reset_data(rgt_dc_params)
            self.rgtPotentialOptions.reset_data(rgt_potentials)
            self.rgtPotentialDirections.reset_data(rgt_pot_dir)
            self.rgtPinnedDirections.reset_data(rgt_pin_dir)
            self.rgtDisplacedDirections.reset_data(rgt_disp_dir)
        except Exception as err:
            logging.exception("Fatal Error: %s", err, extra={'user': 'RGT Logs'})
            self._ctrl.showAlertSignal.emit("Fatal Error", "Error re-loading RGT configurations! Close app and try again.")

    def update_response_params(self, rgt_obj):
        """Update the visible response parameters based on the Response-type."""
        if rgt_obj is None:
            return

        try:
            options_rgt = rgt_obj.configs
            res_type = options_rgt["param_type"]["value"]
            rgt_dc_params = [v for v in options_rgt.values() if v["type"] == "dc-param"]
            for  param in rgt_dc_params:
                if param["id"] == "resistivity":
                    param["visible"] = 1
                else:
                    param["visible"] = res_type
            self.rgtDCParams.reset_data(rgt_dc_params)
        except Exception as err:
            logging.exception("Fatal Error: %s", err, extra={'user': 'RGT Logs'})
            self._ctrl.showAlertSignal.emit("Fatal Error", "Error updating RGT parameters!")

    @Slot()
    def test_method(self):
        print("Test method called!")

    @Slot(result=bool)
    def graph_is_ready(self):
        data_available = self.graph_data_uploaded()
        if not data_available:
            return False
        return self._graph_loaded

    @Slot(result=str)
    def get_pixmap(self):
        """Returns the URL that QML should use to load the image"""
        unique_num = np.random.randint(low=21, high=1000)
        return "image://imageProvider/" + str(unique_num)

    @Slot(result=str)
    def get_file_extensions(self):
        pattern_string = ' '.join(ALLOWED_GRAPH_FILE_EXTENSIONS)
        return f"Graph files ({pattern_string})"

    @Slot(result=bool)
    def enable_edge_list_upload(self):
        if self._ctrl.wait_flag:
            return False
        return not self._ctrl.rgt_obj.edges_uploaded

    @Slot(result=bool)
    def enable_vertex_positions_upload(self):
        if self._ctrl.wait_flag:
            return False
        return not self._ctrl.rgt_obj.vertices_uploaded

    @Slot(result=bool)
    def graph_data_uploaded(self):
        is_edge_list_uploaded = not self.enable_edge_list_upload()
        is_vertex_positions_uploaded = not self.enable_vertex_positions_upload()
        return is_edge_list_uploaded and is_vertex_positions_uploaded

    @Slot(result=bool)
    def is_electrical_response(self):
        response_type = self._ctrl.rgt_obj.configs["response_type"]["value"]
        if response_type == 0:
            return True
        return False

    @Slot(result=bool)
    def is_mechanical_response(self):
        response_type = self._ctrl.rgt_obj.configs["response_type"]["value"]
        if response_type == 1:
            return True
        return False

    @Slot()
    def toggle_response_type(self):
        """Toggle between electrical and mechanical response."""
        self.imageChangedSignal.emit()

    @Slot()
    def apply_changes(self):
        """Retrieve changes made by the user and apply to the response graph."""
        if not self._applying_changes:  # Disallow concurrent changes
            self._applying_changes = True
            self.update_response_params(self._ctrl.rgt_obj)
        self.changeImageSignal.emit()

    @Slot(str, str)
    def apply_imposed_vertices(self, source: str, data: str):
        """Apply imposed vertices to the response graph."""
        try:
            # Convert string '[30, 1], [20, -1]' to 2 numpy arrays: [30, 20] and [1, -1]
            if data != "":
                # Parse string to a Python list
                pairs = ast.literal_eval(data)  # pairs -> [[30, 1], [20, -1]]

                # Convert to NumPy array
                arr = np.asarray(pairs, dtype=float)

                # Split into two arrays
                lst_vert = arr[:, 0]    # [30. 20.]
                lst_pot = arr[:, 1]     # [ 1. -1.]

                # Potential List
                self._ctrl.rgt_obj.list_data["given_potential_list"]["data"] = lst_pot
                self._ctrl.rgt_obj.list_data["given_potential_list"]["type"] = source
                self._ctrl.rgt_obj.list_data["given_potential_list"]["value"] = 1

                # Vertex List
                self._ctrl.rgt_obj.list_data["vertex_list"]["data"] = lst_vert
                self._ctrl.rgt_obj.list_data["vertex_list"]["type"] = source
                self._ctrl.rgt_obj.list_data["vertex_list"]["value"] = 1
            else:
                # Data is blank
                self._ctrl.rgt_obj.list_data["given_potential_list"]["data"] = None
                self._ctrl.rgt_obj.list_data["given_potential_list"]["value"] = 0
                self._ctrl.rgt_obj.list_data["vertex_list"]["data"] = None
                self._ctrl.rgt_obj.list_data["vertex_list"]["value"] = 0
            # Sync models
            self._ctrl.syncModelSignal.emit(self._ctrl.rgt_obj)
        except Exception as err:
            self._ctrl.rgt_obj.list_data["given_potential_list"]["data"] = None
            self._ctrl.rgt_obj.list_data["given_potential_list"]["type"] = "Custom"
            self._ctrl.rgt_obj.list_data["given_potential_list"]["value"] = -1
            # Sync models and refresh image
            self._ctrl.syncModelSignal.emit(self._ctrl.rgt_obj)
            logging.exception("Upload Error: %s", err, extra={'user': 'RGT Logs'})
            self._ctrl.handle_finished(1, False, ["Upload Error", f"Unable to read your data. Try this format for vertex position and its potential: [position-1, potential-1], [position-2, potential-2], ..."])

    @Slot(str)
    def apply_displacement_vector(self, data: str):
        """Apply the displacement vector to the response graph."""
        try:
            # Convert string '[20, 1]' to a numpy array: [20, 1]
            if data != "":
                # Parse string to a Python list
                txt_arr = ast.literal_eval(data)  # txt_arr -> [20, 1]

                # Convert to NumPy array
                arr = np.asarray(txt_arr, dtype=float)

                # Save to displacement_vector
                self._ctrl.rgt_obj.list_data["displacement_vector"]["data"] = arr
                self._ctrl.rgt_obj.list_data["displacement_vector"]["type"] = "Custom"
                self._ctrl.rgt_obj.list_data["displacement_vector"]["value"] = 1
            else:
                # Data is blank
                self._ctrl.rgt_obj.list_data["displacement_vector"]["data"] = None
                self._ctrl.rgt_obj.list_data["displacement_vector"]["value"] = 0
            # Sync models
            self._ctrl.syncModelSignal.emit(self._ctrl.rgt_obj)
        except Exception as err:
            self._ctrl.rgt_obj.list_data["displacement_vector"]["data"] = None
            self._ctrl.rgt_obj.list_data["displacement_vector"]["type"] = "Custom"
            self._ctrl.rgt_obj.list_data["displacement_vector"]["value"] = -1
            # Sync models and refresh image
            self._ctrl.syncModelSignal.emit(self._ctrl.rgt_obj)
            logging.exception("Upload Error: %s", err, extra={'user': 'RGT Logs'})
            self._ctrl.handle_finished(1, False, ["Upload Error", f"Unable to read your data. Try this format for displacement vector: [value-1, value-2]"])


    @Slot(str, str)
    def upload_file_data(self, file_path: str, param_type: str):
        """Upload a CSV file and return True if successful."""
        if self._ctrl.wait_flag:
            logging.info("Please Wait: Another task is running!", extra={'user': 'RGT Logs'})
            self._ctrl.showAlertSignal.emit("Please Wait", "Another task is running!")
            return

        try:
            self.start_task()
            self._ctrl.submit_job(1, "Upload CSV", (file_path, param_type,), True)
        except Exception as err:
            self.stop_task()
            logging.exception("Upload Error: %s", err, extra={'user': 'RGT Logs'})
            self._ctrl.handle_progress_update(ProgressData(type="error", sender="RGT", message=f"Error occurred!"))
            self._ctrl.handle_finished(1, False, ["Upload Error", f"Fatal error while trying to upload {param_type}."])

    @Slot(str)
    def remove_data(self, data_id: str):
        """Remove uploaded data from the ResponseGT object."""
        list_data = self._ctrl.rgt_obj.list_data
        list_data[data_id]["data"] = None
        list_data[data_id]["value"] = 0
        # Sync models and refresh image
        self._ctrl.syncModelSignal.emit(self._ctrl.rgt_obj)
        self.imageChangedSignal.emit()

    @Slot()
    def run_response_analyzer(self):
        """Run the response analyzer and send a progress signal."""

        if self._ctrl.wait_flag:
            logging.info("Please Wait: Another task is running!", extra={'user': 'RGT Logs'})
            self._ctrl.showAlertSignal.emit("Please Wait", "Another task is running!")
            return

        try:
            if (not self._ctrl.rgt_obj.edges_uploaded) or (not self._ctrl.rgt_obj.vertices_uploaded):
                return

            upload_errors = self._ctrl.rgt_obj.verify_uploaded_data
            if upload_errors is not None:
                self._ctrl.handle_finished(1, False, ["File Error", upload_errors])
                return

            self.start_task()
            self._ctrl.submit_job(1, "Compute Response", (self._ctrl.rgt_obj,), True)
        except Exception as err:
            self.stop_task()
            logging.exception("Response Analyzer Error: %s", err, extra={'user': 'RGT Logs'})
            self._ctrl.handle_progress_update(ProgressData(type="error", sender="RGT", message=f"Fatal error occurred!"))
            self._ctrl.handle_finished(1, False, ["Analyzer Error",  "Fatal error while trying to compute ac-response."])

    @Slot()
    def export_response_to_file(self):
        """Export response (edge currents OR vertex potentials) and save as a file."""
        if self._ctrl.wait_flag:
            logging.info("Please Wait: Another task is running!", extra={'user': 'RGT Logs'})
            self._ctrl.showAlertSignal.emit("Please Wait", "Another task is running!")
            return

        try:
            if self._ctrl.rgt_obj.list_data["calculated_vertex_potentials"]["data"] is None or self._ctrl.rgt_obj.list_data["calculated_edge_currents"]["data"] is None:
                self._ctrl.handle_finished(1, False,["Download Error", "Please run the 'Compute Response' first."])
                return

            self.start_task()
            self._ctrl.submit_job(1, "Save Results", (self._ctrl.rgt_obj,), True)
        except Exception as err:
            self.stop_task()
            logging.exception("Download Error: %s", err, extra={'user': 'RGT Logs'})
            self._ctrl.handle_progress_update(ProgressData(type="error", sender="RGT", message=f"Fatal error occurred!"))
            self._ctrl.handle_finished(1, False, ["Download Error", "Fatal error while trying to save results to file."])



# SPDX-License-Identifier: GNU GPL v3

"""
A persistent process worker for running long jobs.
"""


import gc
import threading
from multiprocessing import Process, Queue
from PySide6.QtCore import QObject, Signal


def _worker_loop(job_queue, result_queue):
    """Persistent worker loop inside the subprocess."""
    while True:
        job = job_queue.get()
        if job is None:  # poison pill for shutdown
            break
        func, args = job

        try:
            # --- Attach result_queue to the object that will call _update_progress ---
            # If func is a bound method, its instance is func.__self__
            owner = getattr(func, "__self__", None)
            if owner is not None and hasattr(owner, "progress_queue"):
                # attach the subprocess's result_queue to the unpickled instance
                owner.attach_progress_queue(result_queue)

                # --- Run the job ---
                success, data = func(*args)
                result_queue.put((success, data))
        except Exception as e:
            # result_queue.put((False, str(e)))
            print(f"Worker Loop Exception: {e}")


class ProgressListener(QObject):
    """
    Thread that listens to the multiprocessing.Queue and emits signals into QML UI.
    """
    progress = Signal(object)
    finished = Signal(bool, object)

    def __init__(self, queue: Queue, parent=None):
        super().__init__(parent)
        self._updates_queue = queue
        self._running = True
        self._thread = threading.Thread(target=self._listen, daemon=True)
        self._thread.start()

    def _listen(self):
        while self._running:
            try:
                status, payload = self._updates_queue.get()  # blocking wait
                if status == "STOP":
                    self._running = False
                    break  # Poison pill to stop the thread, otherwise keep running

                if type(status) is str:
                    self.progress.emit(payload)
                else:
                    self.finished.emit(status, payload)
            except Exception as e:
                print(f"Status Thread Listener Exception: {e}")

    def stop(self):
        try:
             self._updates_queue.put_nowait(("STOP", None))  # wakes up the blocking get()
        except Exception as e:
             print(f"Thread Listener Exception: {e}")


class PersistentProcessWorker(QObject):

    # workerStarted = Signal()
    inProgress = Signal(object)
    taskCompleted = Signal(int, bool, object)  # worker-id, success/fail, result (object)

    def __init__(self, worker_id, parent=None):
        """
        Creates a 'multiprocessing Process' to perform long-running tasks in the background without affecting the UI
        thread. This has several advantages: (1) each process is assigned its own Python interpreter and memory - this is
        an advantage for CPU-heavy jobs like image processing; (2) bypasses GIL - meaning that multiple processes can
        truly run in parallel on multiple CPU cores; (3) Processors can be forced to terminate immediately - this is not
        possible with Threads once they start; (4) if a Processor crashes, it does not affect the UI thread.

        The drawbacks are: (1) unlike with Threads, memory is not shared - objects must be serialized/pickled, (2)
        start-up cost is higher than Threads, (3) since memory is not shared, communication is via Queues/Pipes.

        We overcome the start-up cost (which causes sluggish-ness) by using Persistent Processes that are started during
        application launch. To clear and release memory, we restart the Processes after every 3 jobs are completed.

        Args:
            worker_id: The unique ID of the Persistent process worker.
            parent: The parent QObject.
        """
        super().__init__(parent)
        self._worker_id = worker_id
        self._job_queue = None
        self._status_queue = None
        self._process = None
        self._waiting = False
        self._status_listener = None
        self._task_count = 1
        self._start()

    @property
    def status_queue(self):
        return self._status_queue

    @property
    def task_count(self):
        return self._task_count

    def _start(self):
        """Start the worker process and the status listener thread."""
        if self._process is None or not self._process.is_alive():
            # start the persistent process
            self._job_queue = Queue()
            self._status_queue = Queue()
            self._process = Process(target=_worker_loop, args=(self._job_queue, self._status_queue))
            self._process.start()

            if self._status_listener:
                self._status_listener.stop()
            # start a progress/status listener thread
            self._status_listener = ProgressListener(self._status_queue)
            self._status_listener.progress.connect(self.inProgress)
            self._status_listener.finished.connect(self.on_finished)
            # self._status_listener.finished.connect(lambda success, result: self.taskFinishedSignal.emit(self._worker_id, success, result))
            # self.workerStarted.emit()

    def stop(self):
        """Force terminate the worker process."""
        if self._status_listener:
            self._status_listener.stop()
        self._status_listener = None

        if self._process and self._process.is_alive():
            # stop process
            try:
                self._job_queue.put_nowait(None)  # send poison pill
            except Exception as e:
                print(f"Unable to stop, job queue is full: {e}")
            self._process.terminate()
            self._process.join()
        self._process = None
        self._job_queue.close()
        self._status_queue.close()

    """
    # Not effective because it does not release memory - memory is only released when constructor is used to create a
    # new object.
    def restart(self):
        self.workerStarted.connect(lambda : self.on_finished(True, None))
        self.stop()
        self._start()
    """

    def on_finished(self, success, result):
        self._waiting = False
        self.taskCompleted.emit(self._worker_id, success, result)
        # Trigger GC after the job finishes
        gc.collect()
        self._task_count += 1

    def submit_task(self, func, args=()):
        if self._waiting:
            return False  # already busy
        try:
            self._job_queue.put_nowait((func, args))
            self._waiting = True
        except Exception as e:
            print(f"Job queue is full: {e}")
            return False
        return True



# SPDX-License-Identifier: GNU GPL v3

"""
Pyside6 implementation of ResponseGT user interface.
"""

import os
import sys
from PySide6.QtCore import QObject
from PySide6.QtWidgets import QApplication
from PySide6.QtQml import QQmlApplicationEngine

from .controllers.main_controller import MainController
from .models.image_provider import ImageProvider


class PySideApp(QObject):

    def _initialize_models(self):
        """Initialize the models and providers used by the QML engine."""
        self._ui_engine.addImageProvider("imageProvider", self._image_provider)
        self._ui_engine.rootContext().setContextProperty("metricsModel", self._ctrl.network_ctrl.metricsModel)
        self._ui_engine.rootContext().setContextProperty("rgtFiles", self._ctrl.network_ctrl.rgtFiles)
        self._ui_engine.rootContext().setContextProperty("rgtParamOptions", self._ctrl.network_ctrl.rgtParamOptions)
        self._ui_engine.rootContext().setContextProperty("rgtResponseOptions", self._ctrl.network_ctrl.rgtResponseOptions)
        self._ui_engine.rootContext().setContextProperty("rgtDCParams", self._ctrl.network_ctrl.rgtDCParams)
        self._ui_engine.rootContext().setContextProperty("rgtPotentialOptions", self._ctrl.network_ctrl.rgtPotentialOptions)
        self._ui_engine.rootContext().setContextProperty("rgtPotentialDirections", self._ctrl.network_ctrl.rgtPotentialDirections)
        self._ui_engine.rootContext().setContextProperty("rgtPinnedDirections", self._ctrl.network_ctrl.rgtPinnedDirections)
        self._ui_engine.rootContext().setContextProperty("rgtDisplacedDirections", self._ctrl.network_ctrl.rgtDisplacedDirections)

    def _initialize_controllers(self):
        """Initialize the controllers used by the QML engine."""
        self._ui_engine.rootContext().setContextProperty("mainController", self._ctrl)
        self._ui_engine.rootContext().setContextProperty("themeManager", self._ctrl.theme_ctrl)
        self._ui_engine.rootContext().setContextProperty("networkController", self._ctrl.network_ctrl)

    def __init__(self):
        super().__init__()
        self.app = QApplication(sys.argv)
        self._ui_engine = QQmlApplicationEngine()
        self._qml_file = 'qml/MainWindow.qml'

        # IMPORTANT: register the "qml" directory
        qml_dir = os.path.dirname(os.path.abspath(__file__))
        qml_path = os.path.join(qml_dir, self._qml_file)
        self._ui_engine.addImportPath(os.path.join(qml_dir, "qml"))

        ## Register Controller for Dynamic Updates
        self._ctrl = MainController(qml_app=self.app)
        ## Register Image Provider
        self._image_provider = ImageProvider(self._ctrl)

        # Set Models in QML Context
        self._initialize_models()
        self._initialize_controllers()

        ## Cleanup when the app is closing
        self.app.aboutToQuit.connect(self._ctrl.cleanup_workers)

        # Set Theme for the entire application UI ('Basic', 'Fusion', 'Imagine', 'Material', 'Universal')
        # QQuickStyle.setStyle("Basic")

        # Load the QML file and display it
        self._ui_engine.load(qml_path)
        if not self._ui_engine.rootObjects():
            sys.exit(-1)

    @classmethod
    def start(cls) -> None:
        """
        Initialize and run the PySide GUI application.
        """
        gui_app = cls()
        sys.exit(gui_app.app.exec())



# SPDX-License-Identifier: GNU GPL v3
"""
Compute AC response metrics
"""

import re
import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
from sgtlib.modules import ProgressUpdate, plot_to_opencv, ProgressData
from scipy.sparse.linalg import spsolve, lsqr
from scipy.sparse import diags, csc_array, csr_matrix
from matplotlib.collections import LineCollection

from ..utils.config_loader import load_rgt_configs, initialize_list_data

ALLOWED_GRAPH_FILE_EXTENSIONS = ['*.csv']
class ResponseAnalyzer(ProgressUpdate):

    def __init__(self, config_file: str = ""):
        """"""
        super(ResponseAnalyzer, self).__init__()
        self._configs: dict = load_rgt_configs(config_file)
        self._list_data: dict = initialize_list_data()
        self._network_img: None | np.ndarray = None
        self._save_path: str = ""
        self._props: list = []

    @property
    def network_img(self) -> None | np.ndarray:
        return self._network_img

    @property
    def configs(self) -> dict:
        return self._configs

    @configs.setter
    def configs(self, configs) -> None:
        self._configs = configs

    @property
    def list_data(self) -> dict:
        return self._list_data

    @list_data.setter
    def list_data(self, list_params) -> None:
        self._list_data = list_params

    @property
    def props(self) -> list:
        return self._props

    @property
    def vertices_uploaded(self) -> bool:
        """Check if the vertex coordinates were uploaded"""
        is_uploaded = True if self.list_data["vertex_positions"]["value"] == 1 else False
        return is_uploaded

    @property
    def edges_uploaded(self) -> bool:
        """Check if the edge list was uploaded"""
        is_uploaded = True if self.list_data["edge_list"]["value"] == 1 else False
        return is_uploaded

    @property
    def verify_uploaded_data(self) -> str | None:
        """Verify if any of the list data item has errors."""
        list_data = self.list_data
        lst_errors = ""
        for key, item in list_data.items():
            if key == "calculated_vertex_potentials" or key == "calculated_edge_currents":
                continue

            if item["value"] == -1:
                lst_errors += f"{item['text']} has errors. Please re-upload the data!"

            if key == "edge_list" and item["data"] is None:
                lst_errors += f"{item['text']} is missing. Please upload it via a CSV file!"

            if key == "vertex_positions" and item["data"] is None:
                lst_errors += f"{item['text']} is missing. Please upload it via a CSV file!"

        if lst_errors == "":
            return None
        return lst_errors

    def reset(self):
        """Reset all properties"""
        self._props = []
        self.list_data: dict = initialize_list_data()

    def copy_rgt_obj(self, other):
        """Copy attributes from another ResponseAnalyzer object"""
        try:
            self._network_img = other.network_img
            self.list_data["calculated_vertex_potentials"]["data"] = other.list_data["calculated_vertex_potentials"]["data"]
            self.list_data["calculated_vertex_potentials"]["value"] = other.list_data["calculated_vertex_potentials"]["value"]
            self.list_data["calculated_edge_currents"]["data"] = other.list_data["calculated_edge_currents"]["data"]
            self.list_data["calculated_edge_currents"]["value"] = other.list_data["calculated_edge_currents"]["value"]
        except AttributeError:
            return

    def get_filenames(self):
        """
        Splits the image path into file name and image directory.

        Returns:
            filename (str): image file name., output_dir (str): image directory path.
        """
        if self._save_path == "":
            return None, None

        output_dir, filename = os.path.split(self._save_path)
        for ext in ALLOWED_GRAPH_FILE_EXTENSIONS:
            ext = ext.replace('*', '')
            pattern = re.escape(ext) + r'$'
            filename = re.sub(pattern, '', filename)
        return filename, output_dir

    def get_parameter_value(self, key: str):
        """Returns the value of a parameter from the RGT configuration"""
        opt_rgt = self.configs
        coefficient = opt_rgt[key]["value"]
        multiplier = opt_rgt[key]["multiplier"]
        # print(f"{key}: {coefficient} * 10^{multiplier} = {coefficient * 10 ** multiplier}")
        return coefficient * 10 ** multiplier

    def get_response_direction(self) -> str | None:
        opt_rgt = self.configs
        res_dir = None  # Default direction

        for i in range(len(opt_rgt["potential_direction"]["items"])):
            if int(opt_rgt["potential_direction"]["items"][i]["value"]) == 1:
                res_dir = opt_rgt["potential_direction"]["items"][i]["id"]
        return res_dir

    def get_pinned_direction(self) -> str | None:
        opt_rgt = self.configs
        res_dir = 'left'

        for i in range(len(opt_rgt["pinned_direction"]["items"])):
            if int(opt_rgt["pinned_direction"]["items"][i]["value"]) == 1:
                res_dir = opt_rgt["pinned_direction"]["items"][i]["id"]
        return res_dir

    def get_displacement_direction(self) -> str | None:
        opt_rgt = self.configs
        res_dir = 'right'

        for i in range(len(opt_rgt["displacement_direction"]["items"])):
            if int(opt_rgt["displacement_direction"]["items"][i]["value"]) == 1:
                res_dir = opt_rgt["displacement_direction"]["items"][i]["id"]
        return res_dir

    def init_list_data(self, silent: bool) -> bool:
        """
        Compute the list parameters for the response analyzer (electrical and mechanical).

        Args:
             silent (bool): If True, the function will not update the status bar or send status messages

        Returns:
            True if the list parameters are successfully initialized, False otherwise.
        """
        self.update_status(ProgressData(percent=10, sender="RGT",
                                        message=f"Initializing response parameters...")) if not silent else None

        # Check if data is read from CSV
        if self.list_data["vertex_positions"]["data"] is None:
            self.update_status(ProgressData(type="error", sender="RGT",
                                            message=f"Vertex positions are missing! Please upload them via a CSV file.")) if not silent else None
            return False
        if self.list_data["edge_list"]["data"] is None:
            self.update_status(ProgressData(type="error", sender="RGT",
                                            message=f"Edge list is missing! Please upload them via a CSV file.")) if not silent else None
            return False

        # Retrieve configs and all data
        opt_rgt = self.configs
        list_data = self.list_data
        edge_list = list_data["edge_list"]["data"]

        response_type = int(opt_rgt["response_type"]["value"])
        if response_type == 0:
            # Initialize lists (data) for electrical response
            vert_pos = list_data["flipped_vertex_coordinates"]["data"]
            resistivity = self.get_parameter_value("resistivity")
            capacitance = self.get_parameter_value("capacitance")
            inductance = self.get_parameter_value("inductance")
            leak_resistivity = self.get_parameter_value("leak_resistivity")

            if list_data["resistivity_list"]["value"] == 0 or list_data["resistivity_list"]["data"] is None:
                # The array of resistance for each EDGE
                list_data["resistivity_list"]["data"] = resistivity * np.ones(len(edge_list))

            if list_data["inductance_list"]["value"] == 0 or list_data["inductance_list"]["data"] is None:
                # The array of inductance for each EDGE
                list_data["inductance_list"]["data"] = inductance * np.ones(len(edge_list))

            if list_data["given_potential_list"]["value"] == 0 or list_data["given_potential_list"]["data"] is None:
                given_potential_fraction = float(opt_rgt["selected_vertex_proportion"]["value"])
                num_vertices = len(vert_pos)
                num_selected = int(given_potential_fraction * num_vertices)
                list_data["given_potential_list"]["data"] = np.zeros(int(2 * num_selected))

            num_vertices = len(vert_pos)
            num_selected = len(list_data["given_potential_list"]["data"])
            vertex_list = np.ones(num_vertices - num_selected)
            if list_data["capacitance_list"]["value"] == 0 or list_data["capacitance_list"]["data"] is None:
                # The array of capacitance for each NODE that is NOT given an applied potential. nodes are taken to be capacitors connected to a grounded potential (0).
                list_data["capacitance_list"]["data"] = capacitance * vertex_list

            if list_data["leak_resistivity_list"]["value"] == 0 or list_data["leak_resistivity_list"]["data"] is None:
                # The array of resistance between each NODE that is NOT given an applied potential and the ground, a "leakage" resistance.
                list_data["leak_resistivity_list"]["data"] = leak_resistivity * vertex_list
            return True
        elif response_type == 1:
            if list_data["displacement_vector"]["value"] == 0 or list_data["displacement_vector"]["data"] is None:
                list_data["displacement_vector"]["data"] = np.array([1, -1])
            return True
        else:
            return False

    def run_analyzer(self) -> None:
        """Executes functions that will compute AC Response and draw the response graph"""

        if self.abort:
            self.update_status([-1, "Problem encountered while running Response Analyzer.."])
            return

        opt_rgt = self.configs
        response_type = int(opt_rgt["response_type"]["value"])
        if response_type != 0 and response_type != 1:
            self.update_status([-1, "Invalid response type! Please select either 'Electrical' or 'Mechanical' response."])
            self.abort = True
            return
        try:
            if response_type == 0:
                # 1a. Compute AC Response
                _, _ = self.compute_ac_response()
            elif response_type == 1:
                # 1b. Compute AC Response with long edges removed
                _, _ = self.compute_mechanical_response(silent=True)
        except IndexError:
            self.update_status([-1, "One or more vertex positions are out of range! Please re-upload Vertex List."])
            self.abort = True
            return

        # 2. Draw Response Graph
        plt_fig = None
        if response_type == 0:
            plt_fig = self.plot_electrical_response()
        elif response_type == 1:
            plt_fig = self.plot_mechanical_response()
        self.update_status(ProgressData(percent=80, sender="RGT", message=f"Saving graph plot..."))
        self._network_img = plot_to_opencv(plt_fig) if plt_fig is not None else None

    def compute_ac_response(self, silent: bool = False) -> tuple[None, None] | tuple[np.ndarray, np.ndarray]:
        """
        From my testing on square-lattice networks, this method has the time complexity of O(n^~1.4)

        Time complexity might increase significantly in networks with higher average node degree

        :return: potential_response (as a numpy array), current_response (as a numpy array)
        """

        def incidence_matrix():
            """
            Makes an incidence matrix from a list of edges in O(n) time

            :returns: the incidence matrix of the network, where rows are directed edges and columns are vertices
            """

            # We first initialize lists, to which we will append the row, column, and value of the non-zero elements that will fill our sparse incidence array
            c_rows = []
            c_cols = []
            c_vals = []

            # Appending non-zero entries and their row/col data for each edge in the list. Edges are considered directed (+1/-1), but the direction does not matter
            num_edges = len(edge_list)
            num_vertices = len(vert_pos)
            for idx in range(len(edge_list)):
                c_rows.append(idx)
                c_cols.append(int(edge_list[idx, 0]))
                c_vals.append(-1)
                c_rows.append(idx)
                c_cols.append(int(edge_list[idx, 1]))
                c_vals.append(1)
                # It is faster to append each element/coord to a list and then make a sparse array than it is to make a sparse array and then update each element

            incidence_mat = csc_array((c_vals, (c_rows, c_cols)), shape=(num_edges, num_vertices), dtype="complex")
            return incidence_mat

        def make_potential() -> tuple[np.ndarray, np.ndarray]:
            """
            Generate an externally imposed potential field for the system.

            This function constructs a *top-down* potential configuration, where a
            prescribed potential is applied across the system from the top boundary
            to the bottom boundary. Adjusting this function allows you to impose
            different boundary conditions or potential patterns on the same physical
            system, enabling various types of experiments or simulations.

            Returns:
                tuple[np.ndarray, np.ndarray]:
                    - ua_list: Array of potential values applied to boundary nodes.
                    - va_list: Array of node indices corresponding to the applied potentials.
            """
            opt_rgt = self.configs
            given_potential_magnitude = float(opt_rgt["potential_magnitude"]["value"])
            potential_direction = self.get_response_direction()

            # Ultimately this is what gets passed on; the other code in this block just makes this a top-bottom potential
            given_potential_list = list_data["given_potential_list"]["data"]
            vertex_list = np.zeros_like(given_potential_list, dtype=int)
            num_selected = int(len(given_potential_list) / 2)

            if potential_direction == "LR" or potential_direction == "RL":
                # Sort vertices by x-position (Left-Right or Right-Left)
                sorted_vertices = sorted(enumerate(vert_pos), key=lambda v: v[1][0])  # Sort by x-coordinate
                top_vertices = []
                bottom_vertices = []
                left_vertices = sorted_vertices[:num_selected]  # Lowest x-values
                right_vertices = sorted_vertices[-num_selected:]  # Highest x-values
            else:
                # Sort vertices by y-position (Top-Bottom or Bottom-Top)
                sorted_vertices = sorted(enumerate(vert_pos), key=lambda x_pos: x_pos[-1][-1])  # Sort by y-coordinate
                top_vertices = sorted_vertices[-num_selected:]                                  # Top f% (highest y-values)
                bottom_vertices = sorted_vertices[:num_selected]                                # Bottom f% (lowest y-values)
                left_vertices = []
                right_vertices = []

            # Select the potential direction: 'top-down' or 'bottom-up' or 'left-right' or 'right-left'
            if potential_direction == "LR":
                pos_list = [idx for idx, _ in left_vertices]
                neg_list = [idx for idx, _ in right_vertices]
            elif potential_direction == "RL":
                pos_list = [idx for idx, _ in right_vertices]
                neg_list = [idx for idx, _ in left_vertices]
            elif potential_direction == "BT":
                pos_list = [idx for idx, _ in bottom_vertices]
                neg_list = [idx for idx, _ in top_vertices]
            else:
                # Top-Bottom
                pos_list = [idx for idx, _ in top_vertices]
                neg_list = [idx for idx, _ in bottom_vertices]

            # Assign potentials: +V to top
            for idx in range(len(pos_list)):
                given_potential_list[idx] = given_potential_magnitude
                vertex_list[idx] = pos_list[idx]

            # Assign potentials: -V to bottom
            for idx in range(len(neg_list)):
                given_potential_list[len(pos_list) + idx] = -given_potential_magnitude
                vertex_list[len(pos_list) + idx] = neg_list[idx]

            # vertex_list = np.array(vertex_list, dtype=int) # numpy array of vertices that have a forced potential casting to int in case given as float
            return given_potential_list, vertex_list

        self.update_status(ProgressData(percent=1, sender="RGT", message=f"Computing AC response...")) if not silent else None

        # Initialize response parameters and parameter lists
        data_ok = self.init_list_data(silent=silent)
        if not data_ok:
            return None, None
        list_data = self.list_data
        edge_list = list_data["edge_list"]["data"]
        vert_pos = list_data["flipped_vertex_coordinates"]["data"]

        cap_list = list_data["capacitance_list"]["data"]
        leak_res_list = list_data["leak_resistivity_list"]["data"]
        res_list = list_data["resistivity_list"]["data"]
        ind_list = list_data["inductance_list"]["data"]

        # Apply imposing potentials by direction
        self.update_status(ProgressData(percent=5, sender="RGT", message=f"Imposing response potential...")) if not silent else None
        c_mat = incidence_matrix()                          # The incidence matrix of the network, where rows are directed edges and columns are vertices
        if list_data["given_potential_list"]["value"] == 1 and list_data["vertex_list"]["data"] is not None:
            ua_list = list_data["given_potential_list"]["data"]
            va_list = np.array(list_data["vertex_list"]["data"], dtype=int)
        else:
            ua_list, va_list = make_potential()             # ua_list: array applied potentials; va_list: array of nodes with the corresponding applied potential

        given_potential_frequency = self.get_parameter_value("potential_frequency")
        omega = given_potential_frequency                   # The angular frequency of applied alternating potential
        vertices_count = c_mat[0].shape[0]                  # Number of vertices in the graph
        va_vertices_count = int(len(va_list))               # number of vertices in va_list
        vb_vertices_count = int(vertices_count - va_vertices_count)  # number of vertices in vb_list

        # Compute admittance, dynamical and auxiliary matrices
        self.update_status(ProgressData(percent=15, sender="RGT", message=f"Computing admittance, dynamical and auxiliary matrices...")) if not silent else None
        admittance_mat = diags(1 / (res_list + 1j * omega * ind_list))  # diags(1/(rho+1j*omega*inductance)*np.ones(len(edges))) #Y=(R+iwL)^-1, admittance matrix
        c_mat_transposed = c_mat.T                          # transpose of incidence matrix
        dynamical_mat = c_mat_transposed @ admittance_mat @ c_mat  # sparse version of the dynamical matrix

        auxiliary_mat = np.zeros(vertices_count)            # auxiliary array where the value stored at an index is 1 if that index is in va_list
        for vert in va_list:
            auxiliary_mat[vert] = 1
        vb_list = []
        for i in range(vertices_count):  # constructing vb_list in O(n) time
            if auxiliary_mat[i] == 0:
                vb_list.append(i)

        aux_a_index_list = np.zeros(vertices_count)
        for i in range(len(va_list)):
            aux_a_index_list[va_list[i]] = i

        aux_b_index_list = np.zeros(vertices_count)
        for i in range(len(vb_list)):
            aux_b_index_list[vb_list[i]] = i

        dynamical_mat_coo = dynamical_mat.tocoo()
        dynamical_mat_as_list = np.column_stack((dynamical_mat_coo.row, dynamical_mat_coo.col, dynamical_mat_coo.data))

        aa_rows = []
        aa_cols = []
        aa_vals = []

        bb_rows = []
        bb_cols = []
        bb_vals = []

        ba_rows = []
        ba_cols = []
        ba_vals = []

        for elem in dynamical_mat_as_list:
            x = int(np.real(elem[0]))
            y = int(np.real(elem[1]))
            val = elem[2]

            if auxiliary_mat[x] == 1:
                if auxiliary_mat[y] == 1:  # Daa
                    sp_x = int(aux_a_index_list[x])
                    sp_y = int(aux_a_index_list[y])
                    aa_rows.append(sp_x)
                    aa_cols.append(sp_y)
                    aa_vals.append(val)
            else:
                if auxiliary_mat[y] == 0:  # Dbb
                    sp_x = int(aux_b_index_list[x])
                    sp_y = int(aux_b_index_list[y])
                    bb_rows.append(sp_x)
                    bb_cols.append(sp_y)
                    bb_vals.append(val)
                else:  # Dba
                    sp_x = int(aux_b_index_list[x])
                    sp_y = int(aux_a_index_list[y])
                    ba_rows.append(sp_x)
                    ba_cols.append(sp_y)
                    ba_vals.append(val)

        # Creating the sub-matrices for the Dynamical matrix, dynamical_aa, dynamical_bb, dynamical_ba
        # dynamical_aa = csc_array((aa_vals, (aa_rows, aa_cols)), shape=(va_vertices_count, va_vertices_count), dtype="complex")
        dynamical_ba = csc_array((ba_vals, (ba_rows, ba_cols)), shape=(vb_vertices_count, va_vertices_count), dtype="complex")
        dynamical_bb = csc_array((bb_vals, (bb_rows, bb_cols)), shape=(vb_vertices_count, vb_vertices_count), dtype="complex")

        self.update_status(ProgressData(percent=50, sender="RGT", message=f"Solving response equation...")) if not silent else None
        # Need to solve the equation -Dba ua = (Dbb - alpha I) ub
        # LHS = p_mat, RHS = q_mat.ub
        # calculating p_mat and q_mat:
        p_mat = - dynamical_ba @ ua_list
        q_mat = dynamical_bb - diags(1j * omega * cap_list) - diags(1 / leak_res_list)

        # solving for u_b
        ub_list = spsolve(q_mat, p_mat)

        self.update_status(ProgressData(percent=75, sender="RGT", message=f"Computing potential response...")) if not silent else None
        # calculating potential response
        potential_response = np.zeros(vertices_count, dtype="complex")

        # splicing the 'a' and 'b' components of the response back into a single list
        for i in range(len(va_list)):
            potential_response[va_list[i]] = ua_list[i]

        for i in range(len(vb_list)):
            potential_response[vb_list[i]] = ub_list[i]

        self.update_status(ProgressData(percent=85, sender="RGT", message=f"Computing current response...")) if not silent else None
        # calculating current response
        current_response = admittance_mat @ c_mat @ potential_response

        # Save computations
        self.list_data["calculated_vertex_potentials"]["data"] = potential_response
        self.list_data["calculated_vertex_potentials"]["value"] = 1
        self.list_data["calculated_edge_currents"]["data"] = current_response
        self.list_data["calculated_edge_currents"]["value"] = 1
        return potential_response, current_response

    def compute_mechanical_response(self, silent: bool = False) -> tuple[None, None] | tuple[np.ndarray, np.ndarray]:
        """
            Calculates the 2D structural response to static displacements. Uses lsqr to find the minimum-norm solution,
            natively handling singular matrices caused by zero-energy (floppy) modes.

            :return: All displacements (as a Numpy array) and all tensions (as a Numpy array).
        """

        def pinned_compatibility_matrix() -> tuple[csr_matrix, np.ndarray, np.ndarray, np.ndarray]:
            """
                Constructs a pinned 2D compatibility matrix and returns the reduced geometry.
                Pins a specified fraction of vertices based on their spatial ranking.

                :return: A pinned compatibility matrix (as a scipy.sparse.csr_matrix), an array of pinned vertex positions,
                 an array of unpinned edge indices, and a boolean array indicating which edges are valid (not floppy).
            """

            pinned_side = self.get_pinned_direction()
            if pinned_side == 'right':
                use_smallest_boolean = False
                cartesian_direction = 0
            elif pinned_side == 'top':
                use_smallest_boolean = False
                cartesian_direction = 1
            elif pinned_side == 'bottom':
                use_smallest_boolean = True
                cartesian_direction = 1
            else: # left
                use_smallest_boolean = True
                cartesian_direction = 0

            num_vertices = len(vertex_positions)
            num_edges = len(edge_list)

            # 1. Edges by endpoints
            p1 = vertex_positions[edge_list[:, 0]]
            p2 = vertex_positions[edge_list[:, 1]]

            # 2. Construct normal compatibility_mat (2D)
            diff = p1 - p2
            lengths = np.linalg.norm(diff, axis=1, keepdims=True)
            hats = diff / lengths

            rows = np.repeat(np.arange(num_edges), 4)
            cols = np.zeros(num_edges * 4, dtype=int)
            data = np.zeros(num_edges * 4)

            u_idx = edge_list[:, 0]
            v_idx = edge_list[:, 1]

            # 2 DOFs per vertex (x, y)
            cols[0::4] = 2 * u_idx
            cols[1::4] = 2 * u_idx + 1
            cols[2::4] = 2 * v_idx
            cols[3::4] = 2 * v_idx + 1

            data[0::4] = hats[:, 0]
            data[1::4] = hats[:, 1]
            data[2::4] = -hats[:, 0]
            data[3::4] = -hats[:, 1]

            compatibility_mat = csr_matrix((data, (rows, cols)), shape=(num_edges, 2 * num_vertices))

            # 3. Identify pinned and unpinned vertices based on a fraction
            coords = vertex_positions[:, cartesian_direction]
            num_to_pin = int(round(num_vertices * selected_vertex_fraction))
            sorted_indices = np.argsort(coords)

            if use_smallest_boolean:
                pinned_vertices = sorted_indices[:num_to_pin]
            else:
                pinned_vertices = sorted_indices[-num_to_pin:]

            # Get unpinned vertices
            all_vertices = np.arange(num_vertices)
            unpinned_vert_indices = np.setdiff1d(all_vertices, pinned_vertices)
            unpinned_vertex_positions = vertex_positions[unpinned_vert_indices]

            # 4. Construct pinned_cmat
            pinned_dof_arr = []
            for pv in pinned_vertices:
                pinned_dof_arr.extend([2 * pv, 2 * pv + 1])

            mask = np.ones(2 * num_vertices, dtype=bool)
            mask[pinned_dof_arr] = False
            remaining_dofs = np.where(mask)[0]
            pinned_cmat = compatibility_mat[:, remaining_dofs]
            # transposed_cmat = (csr_matrix(pinned_cmat)).T

            # 5. Filter and Re-index Edges
            old_to_new_map = np.full(num_vertices, -1, dtype=int)
            old_to_new_map[unpinned_vert_indices] = np.arange(len(unpinned_vert_indices))

            valid_edge_list_mask = (old_to_new_map[u_idx] != -1) & (old_to_new_map[v_idx] != -1)
            valid_edges_old = edge_list[valid_edge_list_mask]
            unpinned_edge_list = old_to_new_map[valid_edges_old]

            return pinned_cmat, unpinned_vertex_positions, unpinned_edge_list, valid_edge_list_mask

        def get_imposed_displacements(unpinned_vert_positions: np.ndarray) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
            """
                Selects a fraction of boundary vertices and assigns a 2D displacement vector to them.

                :param unpinned_vert_positions: An array of unpinned vertex positions.
                :return: Three sets of arrays: v_list_dof (DOFs of selected vertices), u_list_flat
                (displacement vectors), and displaced vertex positions.
            """
            displacement_side = self.get_displacement_direction()
            displacement_vec = list_data["displacement_vector"]["data"]  # displacement_vector=np.array([0, 10])

            if displacement_side == 'left':
                use_smallest_boolean = True
                cartesian_direction = 0
            elif displacement_side == 'top':
                use_smallest_boolean = False
                cartesian_direction = 1
            elif displacement_side == 'bottom':
                use_smallest_boolean = True
                cartesian_direction = 1
            else: # right
                use_smallest_boolean = False
                cartesian_direction = 0

            num_vertices = len(unpinned_vert_positions)
            num_to_select = int(round(num_vertices * selected_vertex_fraction))

            coords = unpinned_vert_positions[:, cartesian_direction]
            sorted_indices = np.argsort(coords)

            if use_smallest_boolean:
                selected_vertices = sorted_indices[:num_to_select]
            else:
                selected_vertices = sorted_indices[-num_to_select:]

            v_list_dof = []
            for vert in selected_vertices:
                v_list_dof.extend([2 * vert, 2 * vert + 1])

            v_list_dof = np.array(v_list_dof)
            u_list_flat = np.tile(displacement_vec, num_to_select)
            displaced_vertices = unpinned_vert_positions[selected_vertices]

            return v_list_dof, u_list_flat, displaced_vertices

        def generate_2d_spring_constants(k=1.0):
            """
            Generates a list of spring constants for the 2D network. Defaults to 1.0 for all edges.

            :return: A list of spring constants (as a Numpy array) for the 2D network.
            """
            num_edges = len(edge_list)
            return np.full(num_edges, k)

        self.update_status(ProgressData(percent=1, sender="RGT", message=f"Computing Mechanical response...")) if not silent else None

        # Initialize response parameters and parameter lists
        data_ok = self.init_list_data(silent=silent)
        if not data_ok:
            return None, None

        opt_rgt = self.configs
        selected_vertex_fraction = float(opt_rgt["selected_vertex_proportion"]["value"])

        list_data = self.list_data
        edge_list = list_data["edge_list"]["data"]
        vertex_positions = list_data["vertex_positions"]["data"]

        # 1. Constructing the compatibility matrix
        self.update_status(ProgressData(percent=15, sender="RGT", message=f"Constructing compatibility matrix...")) if not silent else None
        pinned_compat_mat, unpinned_vert_pos, unpinned_edge_lst, edge_lst_mask = pinned_compatibility_matrix()

        # 2. Applying the load to network
        self.update_status(ProgressData(percent=30, sender="RGT", message=f"Applying load to network...")) if not silent else None
        v_dof_arr, u_disp_vec, displaced_vert_pos = get_imposed_displacements(unpinned_vert_pos)

        # 3. Computing mechanical response
        self.update_status(ProgressData(percent=50, sender="RGT", message=f"Computing mechanical response...")) if not silent else None
        c_mat = (csr_matrix(pinned_compat_mat)).T
        n_total = c_mat.shape[0]

        # Diagonals of k_list form the stiffness matrix
        k_list = generate_2d_spring_constants()
        stiffness_mat = diags(np.array(k_list), format='csr')
        d_mat = c_mat @ stiffness_mat @ c_mat.T

        # v_dof_arr = np.array(v_dof_arr)
        vb_list = np.setdiff1d(np.arange(n_total), v_dof_arr)

        daa = d_mat[v_dof_arr, :][:, v_dof_arr]
        dba = d_mat[vb_list, :][:, v_dof_arr]
        dbb = d_mat[vb_list, :][:, vb_list]

        # u_disp_vec = u_list  # np.array(u_list)
        p = -dba @ u_disp_vec
        u_b = lsqr(dbb, p)[0]

        dab = dba.T
        f_a = (daa @ u_disp_vec) + (dab @ u_b)

        # Reshape forces to 2D
        self.update_status(ProgressData(percent=75, sender="RGT", message=f"Computing tension and displacements...")) if not silent else None
        f_a_2d = f_a.reshape(-1, 2)
        total_applied_force = np.sum(f_a_2d, axis=0)

        all_displacements = np.zeros(n_total)
        all_displacements[v_dof_arr] = u_disp_vec
        all_displacements[vb_list] = u_b

        # Tensions
        all_tensions = stiffness_mat @ c_mat.T @ all_displacements
        active_tensions = all_tensions[edge_lst_mask]

        # Pinned vertices
        pinned_vert_list = []
        for v in vertex_positions:
            if not np.any(np.all(v == unpinned_vert_pos, axis=1)):
                pinned_vert_list.append(v)
        pinned_vert_pos = np.array(pinned_vert_list)

        self.update_status(ProgressData(percent=85, sender="RGT", message=f"Saving results...")) if not silent else None
        print("Total applied force:", total_applied_force)
        # Save results
        self.list_data["pinned_vertex_positions"]["data"] = pinned_vert_pos
        self.list_data["pinned_vertex_positions"]["value"] = 1
        self.list_data["displaced_vertex_positions"]["data"] = displaced_vert_pos
        self.list_data["displaced_vertex_positions"]["value"] = 1
        self.list_data["unpinned_vertex_positions"]["data"] = unpinned_vert_pos
        self.list_data["unpinned_vertex_positions"]["value"] = 1
        self.list_data["unpinned_edge_list"]["data"] = unpinned_edge_lst
        self.list_data["unpinned_edge_list"]["value"] = 1
        # self.list_data["pinned_cmat"]["data"] = c_mat
        # self.list_data["pinned_cmat"]["value"] = 1
        # self.list_data["displacement_vector"]["data"] = u_disp_vec
        # self.list_data["displacement_vector"]["value"] = 1
        # self.list_data["dof_vertices"]["data"] = v_dof_arr
        # self.list_data["dof_vertices"]["value"] = 1

        self.list_data["calculated_displacements"]["data"] = all_displacements
        self.list_data["calculated_displacements"]["value"] = 1
        self.list_data["calculated_tensions"]["data"] = all_tensions
        self.list_data["calculated_tensions"]["value"] = 1
        self.list_data["calculated_active_tensions"]["data"] = active_tensions
        self.list_data["calculated_active_tensions"]["value"] = 1
        return all_displacements, all_tensions

    def plot_electrical_response(self, graph_type: str = "all", show_current_phase: bool = None, vertex_marker_size: float = None, edge_line_width: float = None, show_color_wheel: bool = None, phase_labels: dict = None) -> None | plt.Figure:
        """
        Draws the graph of an electrical response network.
        """

        if self.list_data["calculated_vertex_potentials"]["data"] is None or self.list_data["calculated_edge_currents"]["data"] is None:
            self.update_status(ProgressData(type="error", sender="RGT", message=f"Response data is missing! Please run the compute_response() method first."))
            return None

        def complex_to_rgb(phases, mags):
            """Convert complex numbers to colors (HSV where H=phase, S=1, V=normalized magnitude)."""
            h_mat = phases / (2 * np.pi)
            s_mat = mags
            v_mat = np.ones_like(h_mat)
            hsv = np.stack([h_mat, s_mat, v_mat], axis=-1)
            return mcolors.hsv_to_rgb(hsv)

        def plot_vertices(marker_size: float = 30):
            """Plot graph vertices only."""
            # Normalize magnitudes for vertices
            pot_mags = np.abs(potential_resp)
            pot_mags_normalized = (pot_mags - pot_mags.min()) / (pot_mags.max() - pot_mags.min() + 1e-10)
            pot_phases = np.angle(potential_resp) % (2 * np.pi)

            # pot_mags_normalized
            vertex_colors = complex_to_rgb(pot_phases, np.array([x for x in pot_mags_normalized]))

            # Plot vertices
            ax.scatter(vert_pos[:, 0], vert_pos[:, 1], c=vertex_colors, s=marker_size * pot_mags_normalized, edgecolors='none', zorder=3)

        def plot_edges(edge_lw: float = 3, use_edge_colors: bool = True, normalize_colors: bool = False):
            """Plot graph edges only."""
            # Normalize magnitudes for edges
            cur_mags = np.abs(current_resp)
            cur_mags_normalized = (cur_mags - cur_mags.min()) / (cur_mags.max() - cur_mags.min() + 1e-10)
            cur_phases = np.angle(current_resp) % (2 * np.pi)

            # cur_mags_normalized
            if normalize_colors:
                edge_colors = complex_to_rgb(cur_phases, np.array([1 for _ in cur_mags_normalized]))
            else:
                edge_colors = complex_to_rgb(np.mod(2 * cur_phases, 2 * np.pi), np.array([x for x in cur_mags_normalized]))

            # Plot edges
            edge_segments = np.array([[vert_pos[int(i)], vert_pos[int(j)]] for i, j in edge_list])
            if use_edge_colors:
                edge_collection = LineCollection(edge_segments, colors=edge_colors, linewidths=edge_lw * cur_mags_normalized)
            else:
                edge_collection = LineCollection(edge_segments, colors="0", linewidths=edge_lw * cur_mags_normalized)
            ax.add_collection(edge_collection)

        def plot_color_wheel(phase_angle_labels: dict = None):
            """Plot color wheel."""
            # Create the color-wheel legend
            ax_color = fig.add_axes((0.75, 0.15, 0.15, 0.15), projection='polar')  # this vector is x-position, y-position, width, height of color-wheel

            # Create the color-wheel using bars instead of fill_between
            n_angles = 360
            theta = np.linspace(0, 2 * np.pi, n_angles, endpoint=False)
            width = 2 * np.pi / n_angles
            radii = np.ones(n_angles)

            # Create the HSV colors for the wheel
            h = theta / (2 * np.pi)
            s = np.ones_like(h)
            v = np.ones_like(h)
            hsv = np.stack([h, s, v], axis=-1)
            colors = mcolors.hsv_to_rgb(hsv)

            # Plot the color wheel using bar plot
            ax_color.bar(theta, radii, width=width, bottom=0, color=colors) # uncomment to show color-wheel

            # Add magnitude indicator (radius)
            ax_color.plot([0, 0], [0, 1], 'k-', lw=1)  # magnitude indicator
            ax_color.text(0, 1.1, '|z|', ha='center', va='center')

            # Add the phase-angle labels
            if phase_angle_labels is None:
                phase_angle_labels = {
                    0: '0',
                    np.pi/2: r'$\pi/2$',
                    np.pi: r'$\pi$',
                    3*np.pi/2: r'$3\pi/2$'
                }

            for angle, label in phase_angle_labels.items():
                ax_color.text(angle, 1.3, label, ha='center', va='center')

            ax_color.set_xticks([])
            ax_color.set_yticks([])
            ax_color.set_theta_direction(-1)
            ax_color.set_theta_offset(np.pi / 2)
            ax_color.set_frame_on(False)

        self.update_status(ProgressData(percent=1, sender="RGT", message=f"Generating response graph..."))
        # Retrieve computed response data (should be numpy arrays)
        vert_pos = self.list_data["flipped_vertex_coordinates"]["data"]
        edge_list = self.list_data["edge_list"]["data"]
        potential_resp = self.list_data["calculated_vertex_potentials"]["data"]
        current_resp = self.list_data["calculated_edge_currents"]["data"]

        self.update_status(ProgressData(percent=5, sender="RGT", message=f"Fetching parameters..."))
        # Set up a figure and an axis
        if graph_type == "all":
            fig = plt.Figure(figsize=(16.4 / 2, 10.7 / 2)) # the last ordered pair is the aspect ratio
            ax = fig.add_axes((0, 0, 1, 1))  # span the whole figure
            mk_size = 10 if vertex_marker_size is None else vertex_marker_size
            lw = 5 if edge_line_width is None else edge_line_width
            color_phases = False if show_current_phase is None else show_current_phase

            self.update_status(ProgressData(percent=10, sender="RGT", message=f"Plotting graph vertices..."))
            plot_vertices(marker_size=mk_size)

            self.update_status(ProgressData(percent=40, sender="RGT", message=f"Plotting graph edges..."))
            plot_edges(edge_lw=lw, use_edge_colors=color_phases, normalize_colors=True)
        else:
            fig = plt.Figure(figsize=(9, 9))
            ax = fig.add_axes((0, 0, 1, 1))  # span the whole figure
            if graph_type == "vertices":
                self.update_status(ProgressData(percent=10, sender="RGT", message=f"Plotting graph vertices..."))
                mk_size = 30 if vertex_marker_size is None else vertex_marker_size
                plot_vertices(marker_size=mk_size)
            elif graph_type == "edges":
                self.update_status(ProgressData(percent=10, sender="RGT", message=f"Plotting graph edges..."))
                lw = 3 if edge_line_width is None else edge_line_width
                color_phases = True if show_current_phase is None else show_current_phase
                plot_edges(edge_lw=lw, use_edge_colors=color_phases)

        if show_color_wheel:
            self.update_status(ProgressData(percent=60, sender="RGT", message=f"Plotting color wheel..."))
            plot_color_wheel(phase_labels)

        # Determine plot ranges
        x_min, x_max = vert_pos[:, 0].min(), vert_pos[:, 0].max()
        y_min, y_max = vert_pos[:, 1].min(), vert_pos[:, 1].max()
        x_pad = (x_max - x_min) * 0.1
        y_pad = (y_max - y_min) * 0.1
        ax.set_xlim(x_min - x_pad, x_max + x_pad)
        ax.set_ylim(y_min - y_pad, y_max + y_pad)
        #ax.set_axis_off()

        # fig.tight_layout()
        return fig

    def plot_mechanical_response(self) -> None | plt.Figure:
        """
        Draws the graph of a mechanical response network.
        """
        if self.list_data["calculated_displacements"]["data"] is None or self.list_data["calculated_tensions"]["data"] is None\
                or self.list_data["unpinned_vertex_positions"]["data"] is None\
                or self.list_data["pinned_vertex_positions"]["data"] is None\
                or self.list_data["displaced_vertex_positions"]["data"] is None:
            self.update_status(ProgressData(type="error", sender="RGT", message=f"Response data is missing! Please run the compute_response() method first."))
            return None

        # Fetch calculated data
        self.update_status(ProgressData(percent=1, sender="RGT", message=f"Generating response graph..."))
        edge_list = self.list_data["edge_list"]["data"]
        vert_pos = self.list_data["vertex_positions"]["data"]
        displaced_vert_pos = self.list_data["displaced_vertex_positions"]["data"]
        pinned_vertices = self.list_data["pinned_vertex_positions"]["data"]
        unpinned_vertices = self.list_data["unpinned_vertex_positions"]["data"]
        #unpinned_edges = self.list_data["unpinned_edge_list"]["data"]
        all_tensions = self.list_data["calculated_tensions"]["data"]
        #active_tensions = self.list_data["calculated_active_tensions"]["data"]
        all_displacements = self.list_data["calculated_displacements"]["data"]
        zero_mask_threshold = 1e-8
        moved_mask_threshold = 1e-10

        # --- SET UP the figure, and axis ---
        fig = plt.figure(figsize=(10, 10))
        ax = fig.add_axes((0, 0.08, 1, 0.92))   # whole figure with a small bottom margin
        ##fig, ax = plt.subplots(figsize=(10, 10))

        # --- EDGES (TENSION VISUALIZATION FOR ALL EDGES) ---
        self.update_status(ProgressData(percent=10, sender="RGT", message=f"Plotting unpinned edges..."))
        all_segments = vert_pos[edge_list]

        # --- MASK for tension edges ---
        zero_mask = np.abs(all_tensions) < zero_mask_threshold
        nonzero_mask = ~zero_mask

        # Add zero-tension edges
        if np.any(zero_mask):
            lc_zero = LineCollection(all_segments[zero_mask], colors='gray', linewidths=0.5, linestyles='dashed', alpha=0.5, zorder=2, label='Zero Tension')
            ax.add_collection(lc_zero)

        # Add non-zero-tension edges
        if np.any(nonzero_mask):
            max_t = np.max(np.abs(all_tensions[nonzero_mask]))
            normalized_tensions = np.abs(all_tensions[nonzero_mask]) / (max_t + 1e-12)
            dynamic_linewidths = 0.5 + (4.0 * normalized_tensions)
            lc_nonzero = LineCollection(all_segments[nonzero_mask], colors='black', linewidths=dynamic_linewidths, alpha=0.8, zorder=3, label='Tensioned Edges')
            ax.add_collection(lc_nonzero)

        """
        # --- EDGES (TENSION VISUALIZATION) ---
        self.update_status(ProgressData(percent=20, sender="RGT", message=f"Plotting unpinned edges..."))
        segments = unpinned_vertices[unpinned_edges]

        zero_mask = np.abs(active_tensions) < zero_mask_threshold
        nonzero_mask = ~zero_mask

        # Add zero-tension edges
        if np.any(zero_mask):
            lc_zero = LineCollection(segments[zero_mask], colors='gray', linewidths=0.5, linestyles='dashed', alpha=0.5, zorder=2)
            ax.add_collection(lc_zero)

        # Add non-zero-tension edges
        if np.any(nonzero_mask):
            max_t = np.max(np.abs(active_tensions[nonzero_mask]))
            normalized_tensions = np.abs(active_tensions[nonzero_mask]) / (max_t + 1e-12)
            dynamic_linewidths = 0.5 + (4.0 * normalized_tensions)
            lc_nonzero = LineCollection(segments[nonzero_mask], colors='black', linewidths=dynamic_linewidths, alpha=0.8, zorder=3)
            ax.add_collection(lc_nonzero)
        """

        # --- VERTICES ---
        self.update_status(ProgressData(percent=20, sender="RGT", message=f"Plotting graph vertices..."))
        # Plot all original vertices as the background
        ax.scatter(vert_pos[:, 0], vert_pos[:, 1], color='lightgray', s=10, alpha=0.5, zorder=1)

        # Plot active unpinned vertices darker
        x_unp = unpinned_vertices[:, 0]
        y_unp = unpinned_vertices[:, 1]
        ax.scatter(x_unp, y_unp, color='black', s=15, alpha=0.9, zorder=4, label='Free Vertices')

        # Plot pinned vertices
        if len(pinned_vertices) > 0:
            ax.scatter(pinned_vertices[:, 0], pinned_vertices[:, 1], marker='s', color='black', s=40, zorder=5, label='Pinned Vertices')

        # Highlight vertices with applied displacements
        ax.scatter(displaced_vert_pos[:, 0], displaced_vert_pos[:, 1], facecolors='none', edgecolors='blue', s=100, linewidths=2, zorder=6, label='Applied Displacement')

        # --- DISPLACEMENT ARROWS ---
        self.update_status(ProgressData(percent=50, sender="RGT", message=f"Plotting displacement arrows..."))
        # Reshape to 2D
        disp_2d = all_displacements.reshape(-1, 2)
        u_vec, v_vec = disp_2d[:, 0], disp_2d[:, 1]
        disp_magnitudes = np.linalg.norm(disp_2d, axis=1)

        cmap = plt.get_cmap('turbo')
        norm = mcolors.Normalize(vmin=disp_magnitudes.min(), vmax=disp_magnitudes.max())
        arrow_colors = cmap(norm(disp_magnitudes))
        moved_mask = disp_magnitudes > moved_mask_threshold

        if np.any(moved_mask):
            # Auto-scaling arrows to sensible dimensions for readability. Note: angles='xy' ensures vectors correctly map to graph geometry
            ax.quiver(x_unp[moved_mask], y_unp[moved_mask], u_vec[moved_mask], v_vec[moved_mask], color=arrow_colors[moved_mask], angles='xy', zorder=7)

        # --- COLORBAR & LEGEND ---
        self.update_status(ProgressData(percent=75, sender="RGT", message=f"Adding colorbar..."))
        sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
        sm.set_array([])
        cbar = fig.colorbar(sm, ax=ax, shrink=0.5, pad=0.05)
        cbar.set_label('Displacement Magnitude')

        handles, labels = ax.get_legend_handles_labels()
        by_label = dict(zip(labels, handles))
        #ax.legend(by_label.values(), by_label.keys(), loc='upper right', bbox_to_anchor=(1.35, 1))
        fig.legend(by_label.values(), by_label.keys(), loc="upper center", bbox_to_anchor=(0.45, 0.25), ncol=3)

        # --- AESTHETICS & SCALING ---
        ax.set_title("Network Topology and Static Response")
        ax.set_xlabel("X")
        ax.set_ylabel("Y")
        ax.set_aspect('equal', adjustable='box')  # Keeps geometry undistorted

        buffer = 0.1
        x_range = vert_pos[:, 0].max() - vert_pos[:, 0].min()
        y_range = vert_pos[:, 1].max() - vert_pos[:, 1].min()

        ax.set_xlim(vert_pos[:, 0].min() - buffer * x_range, vert_pos[:, 0].max() + buffer * x_range)
        ax.set_ylim(vert_pos[:, 1].min() - buffer * y_range, vert_pos[:, 1].max() + buffer * y_range)

        #fig.tight_layout()
        #plt.show()
        return fig

    def save_results_to_file(self) -> bool:
        """Save computed response data to a file."""
        if self.list_data["calculated_vertex_potentials"]["data"] is None or self.list_data["calculated_edge_currents"]["data"] is None:
            return False

        filename, out_dir = self.get_filenames()
        if filename is None:
            return False

        edges_filename = filename + "_EdgeCurrents.csv"
        nodes_filename = filename + "_VertexPotentials.csv"
        edges_file = os.path.join(out_dir, edges_filename)
        nodes_file = os.path.join(out_dir, nodes_filename)

        # Export the DataFrame to a CSV file without header and index
        df_pot = pd.DataFrame(self.list_data["calculated_vertex_potentials"]["data"])
        df_pot.to_csv(str(nodes_file), header=False, index=False)

        # Export the DataFrame to a CSV file without header and index
        df_curr = pd.DataFrame(self.list_data["calculated_edge_currents"]["data"])
        df_curr.to_csv(str(edges_file), header=False, index=False)
        return True


from PySide6.QtCore import Qt, QAbstractListModel, QModelIndex, QPersistentModelIndex

# Define a simple QAbstractListModel
class CheckBoxModel(QAbstractListModel):
    IdRole = Qt.ItemDataRole.UserRole + 1
    TypeRole = Qt.ItemDataRole.UserRole + 2
    TextRole = Qt.ItemDataRole.UserRole + 3
    ValueRole = Qt.ItemDataRole.UserRole + 4
    DataIdRole = Qt.ItemDataRole.UserRole + 5
    DataValueRole = Qt.ItemDataRole.UserRole + 6
    MinValueRole = Qt.ItemDataRole.UserRole + 7
    MaxValueRole = Qt.ItemDataRole.UserRole + 8
    StepSizeRole = Qt.ItemDataRole.UserRole + 9
    VisibleRole = Qt.ItemDataRole.UserRole + 10
    ToolTipRole = Qt.ItemDataRole.ToolTipRole + 11
    MultiplierRole = Qt.ItemDataRole.UserRole + 12

    def __init__(self, data: list, parent=None):
        super().__init__(parent)
        self.list_data = data

    def rowCount(self, parent=None):
        return len(self.list_data)

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        if not index.isValid() or index.row() >= len(self.list_data):
            return None
        item = self.list_data[index.row()]

        # Map roles → dictionary keys
        role_map = {
            self.IdRole: "id",
            self.TypeRole: "type",
            self.TextRole: "text",
            self.ValueRole: "value",
            self.DataIdRole: "dataId",
            self.DataValueRole: "dataValue",
            self.MinValueRole: "minValue",
            self.MaxValueRole: "maxValue",
            self.StepSizeRole: "stepSize",
            self.VisibleRole: "visible",
            self.ToolTipRole: "tooltip",
            self.MultiplierRole: "multiplier",
        }

        key = role_map.get(role, None)
        if key is None:
            return None

        # Safe return: return value if key exists, else None
        return item.get(key, None)

    def setData(self, index, value, role=Qt.ItemDataRole.DisplayRole):
        """

        Args:
            index (QModelIndex | QPersistentModelIndex):
            value (int|float):
            role (int):
        """

        if not index.isValid() or index.row() >= len(self.list_data):
            return False

        if role == self.ValueRole:
            self.list_data[index.row()]["value"] = value
            self.dataChanged.emit(index, index, [role])
            return True
        if role == self.DataValueRole:
            self.list_data[index.row()]["dataValue"] = value
            self.dataChanged.emit(index, index, [role])
            return True
        if role == self.MultiplierRole:
            self.list_data[index.row()]["multiplier"] = value
            self.dataChanged.emit(index, index, [role])
            return True
        return False

    def reset_data(self, new_data: list):
        self.beginResetModel()
        self.list_data = new_data
        self.endResetModel()
        self.dataChanged.emit(self.index(0,0), self.index(len(new_data), 0),
                              [self.IdRole, self.TypeRole, self.TextRole, self.ValueRole, self.DataIdRole,
                               self.DataValueRole, self.MinValueRole, self.MaxValueRole, self.StepSizeRole,
                               self.VisibleRole, self.ToolTipRole, self.MultiplierRole])

    def roleNames(self):
        return {
            self.IdRole: b"id",
            self.TypeRole: b"type",
            self.TextRole: b"text",
            self.ValueRole: b"value",
            self.DataIdRole: b"dataId",
            self.DataValueRole: b"dataValue",
            self.MinValueRole: b"minValue",
            self.MaxValueRole: b"maxValue",
            self.StepSizeRole: b"stepSize",
            self.VisibleRole: b"visible",
            self.ToolTipRole: b"tooltip",
            self.MultiplierRole: b"multiplier"
        }


# SPDX-License-Identifier: GNU GPL v3

"""
Loads default configurations from 'configs.ini' file
"""

import os
import math
import configparser
import numpy as np
from typing import Union
from sgtlib.modules import verify_path


def read_config_file(config_path):
    """Read the contents of the 'configs.ini' file"""
    config = configparser.ConfigParser()
    success, result = verify_path(config_path)
    if success:
        config_file = result
    else:
        # Using the default config file. Get the directory of the current script
        script_dir = os.path.dirname(os.path.abspath(__file__))
        config_path = 'configs.ini'
        config_file = os.path.join(script_dir, config_path)
    # Load the default configuration from the file
    try:
        config.read(config_file)
        # print(f"Configs loaded successfully from: {config_file}.")
        return config
    except configparser.Error:
        # print(f"Unable to read the configs from {config_file}.")
        return None


def initialize_list_data():
    """Initialize the list of parameters for the ResponseGT computation."""

    response_file_options: dict[str, dict[str, str | int | None | np.ndarray]] = {
        # Electrical Response
        "vertex_positions": {"id": "vertex_positions", "value": 0, "data": None, "type": "File", "text": "Vertex Positions", "visible": 1},
        "edge_list": {"id": "edge_list", "value": 0, "data": None, "type": "File", "text": "Edge List", "visible": 1},
        "resistivity_list": {"id": "resistivity_list", "value": 0, "data": None, "type": "File", "text": "Resistivity List", "visible": 1},
        "inductance_list": {"id": "inductance_list", "value": 0, "data": None, "type": "File", "text": "Inductance List", "visible": 1},
        "capacitance_list": {"id": "capacitance_list", "value": 0, "data": None, "type": "File", "text": "Capacitance List", "visible": 1},
        "leak_resistivity_list": {"id": "leak_resistivity_list", "value": 0, "data": None, "type": "File", "text": "Leak Resistivity List", "visible": 1},
        "given_potential_list": {"id": "given_potential_list", "value": 0, "data": None, "type": "File", "text": "Given Potential List", "visible": 1},
        "vertex_list": {"id": "vertex_list", "value": 0, "data": None, "type": "File", "text": "Vertex List", "visible": 1},
        "flipped_vertex_coordinates": {"id": "flipped_vertex_coordinates", "value": 0, "data": None, "type": "Custom", "text": "Flipped Vertex Coordinates", "visible": 0},
        "calculated_vertex_potentials": {"id": "calculated_vertex_potentials", "value": 0, "data": None, "type": "Custom", "text": "Calculated Vertex Potentials", "visible": 0},
        "calculated_edge_currents": {"id": "calculated_edge_currents", "value": 0, "data": None, "type": "Custom", "text": "Calculated Edge Currents", "visible": 0},
        # Mechanical Response
        "displacement_vector": {"id": "displacement_vector", "value": 0, "data": None, "type": "File", "text": "Displacement Vector", "visible": 1},
        "edge_mask_list": {"id": "edge_mask_list", "value": 0, "data": None, "type": "File", "text": "Valid Edge Mask", "visible": 1},
        "displaced_vertex_positions": {"id": "displaced_vertex_positions", "value": 0, "data": None, "type": "Custom", "text": "Calculated Displaced Vertex Positions", "visible": 0},
        "pinned_vertex_positions": {"id": "pinned_vertex_positions", "value": 0, "data": None, "type": "Custom", "text": "Calculated Pinned Vertex Positions", "visible": 0},
        "unpinned_vertex_positions": {"id": "unpinned_vertex_positions", "value": 0, "data": None, "type": "Custom", "text": "Calculated Unpinned Vertex Positions", "visible": 0},
        "unpinned_edge_list": {"id": "unpinned_edge_list", "value": 0, "data": None, "type": "Custom", "text": "Calculated Unpinned Edge List", "visible": 0},
        "calculated_displacements": {"id": "calculated_displacements", "value": 0, "data": None, "type": "Custom", "text": "Calculated Displacements", "visible": 0},
        "calculated_tensions": {"id": "calculated_tensions", "value": 0, "data": None, "type": "Custom", "text": "Calculated Tensions", "visible": 0},
        "calculated_active_tensions": {"id": "calculated_active_tensions", "value": 0, "data": None, "type": "Custom", "text": "Calculated Active Tensions", "visible": 0},
    }
    return response_file_options


def get_metric_options() -> list[dict[str, Union[str, int]]]:
    """Return a list of metric options for the QML adapter."""
    metric_options = [
        {"text": "10⁻²⁴", "multiplier": -24},
        {"text": "10⁻²¹", "multiplier": -21},
        {"text": "10⁻¹⁸", "multiplier": -18},
        {"text": "10⁻¹⁵", "multiplier": -15},
        {"text": "10⁻¹²", "multiplier": -12},  # picometer
        {"text": "10⁻⁹", "multiplier": -9},  # nanometer
        {"text": "10⁻⁶", "multiplier": -6},  # micrometer
        {"text": "10⁻³", "multiplier": -3},  # millimeter
        {"text": "10⁻²", "multiplier": -2},  # centimeter
        # {"text": "10⁻¹", "multiplier": -1},    # decimeter
        {"text": "10⁰", "multiplier": 0},  # meter
        # {"text": "10¹", "multiplier": 1},       # dekameter
        # {"text": "10²", "multiplier": 2},       # hectometer
        {"text": "10³", "multiplier": 3},  # kilometer
        {"text": "10⁶", "multiplier": 6},  # megameter
        {"text": "10⁹", "multiplier": 9},  # gigameter
        {"text": "10¹²", "multiplier": 12},  # terameter
        {"text": "10¹⁵", "multiplier": 15},
        {"text": "10¹⁸", "multiplier": 18},
        {"text": "10²¹", "multiplier": 21},
        {"text": "10²⁴", "multiplier": 24},
    ]
    return metric_options


def load_rgt_configs(cfg_path: str = ""):
    """ResponseGT computation configuration loader."""

    def number_to_scientific_parts(value: float):
        """
        Convert a number into (coefficient, multiplier) such that:
            value = coefficient * 10**multiplier

        Example:
            0.005 → (5, -3)
            123 → (1.23, 2)

        Special rule:
            If the multiplier is -1, 1, or 2 → force it to 0
        """

        if value == 0:
            return 0, 0

        # Determine scientific exponent
        multiplier = int(math.floor(math.log10(abs(value))))
        coefficient = float(value / (10 ** multiplier))

        # Ensure coefficient is in [1, 10)
        if abs(coefficient) < 1:
            coefficient *= 10
            multiplier -= 1

        # Force multiplier -1, 1, or 2 back to 0
        if multiplier in {-1, 1, 2}:
            coefficient = value  # collapse scaling into coefficient
            multiplier = 0

        lst_mult_data = get_metric_options()
        mult_range = {x["multiplier"] for x in lst_mult_data}
        if multiplier not in mult_range:
            # Find the closest multiplier in the list
            nearest_multiplier = min(mult_range, key=lambda x: abs(x - multiplier))
            # Update the coefficient and multiplier
            coefficient *= 10 ** (multiplier - nearest_multiplier)
            multiplier = nearest_multiplier
        return coefficient, multiplier

    # add the imposed direction (selected)
    options_rgt: dict[str, dict[str, Union[int, float, list]]] = {
        "response_type": {"id": "response_type", "type": "rgt-settings", "text": "Response Type", "visible": 1, "value": 0},
        "param_type": {"id": "param_type", "type": "param-settings", "text": "Response Type", "visible": 1, "value": 0},

        "potential_direction": {"id": "potential_direction", "type": "potential-settings", "text": "Potential Direction", "visible": 0, "value": 1,
                                "items": [
                                    {"id": "TB", "text": "Top-Bottom", "value": 1},
                                    {"id": "BT", "text": "Bottom-Top", "value": 0},
                                    {"id": "LR", "text": "Left-Right", "value": 0},
                                    {"id": "RL", "text": "Right-Left", "value": 0}
                                ]},
        "potential_magnitude": {"id": "potential_magnitude", "type": "potential-settings", "text": "Potential Magnitude", "visible": 1, "value": 100.0, "minValue": -100, "maxValue": 100},
        "selected_vertex_proportion": {"id": "selected_vertex_proportion", "type": "potential-settings", "text": "Fraction of Vertices", "visible": 1, "value": 0.05, "minValue": 0, "maxValue": 1},
        "potential_frequency": {"id": "potential_frequency", "type": "dc-param", "text": "Potential Frequency", "visible": 1, "value": 1, "multiplier": -6, "minValue": -1000, "maxValue": 1000},

        "resistivity": {"id": "resistivity", "type": "dc-param", "text": "Resistivity", "visible": 1, "value": 1, "multiplier": 0, "minValue": -1000, "maxValue": 1000},
        "capacitance": {"id": "capacitance", "type": "dc-param", "text": "Capacitance", "visible": 1, "value": 1, "multiplier": -6, "minValue": -1000, "maxValue": 1000},
        "inductance": {"id": "inductance", "type": "dc-param", "text": "Inductance", "visible": 1, "value": 1, "multiplier": -9, "minValue": -1000, "maxValue": 1000},
        "leak_resistivity": {"id": "leak_resistivity", "type": "dc-param", "text": "Leak Resistivity", "visible": 1, "value": 1, "multiplier": 6, "minValue": -1000, "maxValue": 1000},

        "pinned_direction": {"id": "pinned_direction", "type": "mech-param", "text": "Pinned Side", "visible": 0, "value": 1,
                        "items": [
                            {"id": "left", "text": "Left", "value": 1},
                            {"id": "right", "text": "Right", "value": 0},
                            {"id": "top", "text": "Top", "value": 0},
                            {"id": "bottom", "text": "Bottom", "value": 0}
                        ]},
        "displacement_direction": {"id": "displacement_direction", "type": "mech-param", "text": "Displacement Side", "visible": 0, "value": 1,
                              "items": [
                                  {"id": "left", "text": "Left", "value": 0},
                                  {"id": "right", "text": "Right", "value": 1},
                                  {"id": "top", "text": "Top", "value": 0},
                                  {"id": "bottom", "text": "Bottom", "value": 0}
                              ]},
    }

    # Load configuration from the file
    config = read_config_file(cfg_path)
    if config is None:
        return options_rgt

    try:
        options_rgt["response_type"]["value"] = int(config.get('rgt-settings', 'response_type'))
        options_rgt["param_type"]["value"] = int(config.get('rgt-settings', 'param_type'))
        frac_val = float(config.get('rgt-settings', 'selected_vertex_proportion'))

        pot_dir = str(config.get('dc-response', 'potential_direction'))
        freq_val = float(config.get('dc-response', 'potential_frequency'))
        mag_val = float(config.get('dc-response', 'potential_magnitude'))
        res_val = float(config.get('dc-response', 'resistivity'))
        cap_val = float(config.get('dc-response', 'capacitance'))
        ind_val = float(config.get('dc-response', 'inductance'))
        leak_val = float(config.get('dc-response', 'leak_resistivity'))

        pin_sd = str(config.get('mechanical-response', 'pinned_direction'))
        disp_sd = str(config.get('mechanical-response', 'displacement_direction'))

        options_rgt["selected_vertex_proportion"]["value"] = frac_val
        options_rgt["potential_magnitude"]["value"] = mag_val
        for i in range(len(options_rgt["potential_direction"]["items"])):
            options_rgt["potential_direction"]["items"][i]["value"] = 1 if options_rgt["potential_direction"]["items"][i]["id"] == pot_dir else 0

        options_rgt["potential_frequency"]["value"], options_rgt["potential_frequency"]["multiplier"] = number_to_scientific_parts(freq_val)
        options_rgt["resistivity"]["value"], options_rgt["resistivity"]["multiplier"] = number_to_scientific_parts(res_val)
        options_rgt["capacitance"]["value"], options_rgt["capacitance"]["multiplier"] = number_to_scientific_parts(cap_val)
        options_rgt["inductance"]["value"], options_rgt["inductance"]["multiplier"] = number_to_scientific_parts(ind_val)
        options_rgt["leak_resistivity"]["value"], options_rgt["leak_resistivity"]["multiplier"] = number_to_scientific_parts(leak_val)

        for i in range(len(options_rgt["pinned_direction"]["items"])):
            options_rgt["pinned_direction"]["items"][i]["value"] = 1 if options_rgt["pinned_direction"]["items"][i]["id"] == pin_sd else 0

        for i in range(len(options_rgt["displacement_direction"]["items"])):
            options_rgt["displacement_direction"]["items"][i]["value"] = 1 if options_rgt["displacement_direction"]["items"][i]["id"] == disp_sd else 0

        return options_rgt
    except configparser.NoSectionError:
        return options_rgt



# SPDX-License-Identifier: GNU GPL v3
"""
Checks OS-level theme
"""

from PySide6.QtCore import QObject, Property, Signal, Slot
from PySide6.QtGui import QPalette
from PySide6.QtWidgets import QApplication


class ThemeManager(QObject):
    themeChanged = Signal()

    def __init__(self):
        super().__init__()
        self._isDark = ThemeManager.detect_dark_mode()

    @Property(bool, notify=themeChanged)
    def is_dark(self):
        return self._isDark

    @Slot()
    def reload_theme(self):
        """Call this if the OS theme changes (Windows 11 sends events)."""
        new_state = ThemeManager.detect_dark_mode()
        if new_state != self._isDark:
            self._isDark = new_state
            self.themeChanged.emit()

    # Detect OS theme
    @staticmethod
    def detect_dark_mode():
        palette = QApplication.palette()
        # simple but works on Windows/macOS/Linux
        return palette.color(QPalette.ColorRole.Window).value() < 128



# SPDX-License-Identifier: GNU GPL v3

"""
Base worker class for executing all resource-intensive StructuralGT tasks.
"""

import os
import logging
from sgtlib.modules import ProgressData, TaskResult, verify_path, csv_to_numpy
from ...compute.response_analyzer import ALLOWED_GRAPH_FILE_EXTENSIONS


class BaseWorker:

    def __init__(self):
        self._progress_queue = None

    @property
    def progress_queue(self):
        return self._progress_queue

    def _update_progress(self, status_data: ProgressData):
        """
        Send the update_progress signal to all listeners.
        Progress-value (0-100), progress-message (str)
        Args:
            status_data: ProgressData object that contains the percentage and status message of the current task.

        Returns:

        """
        if self._progress_queue is None:
            return
        self._progress_queue.put(("progress", status_data))

    def attach_progress_queue(self, queue):
        """Attach or replace the progress queue (status_queue)."""
        if self._progress_queue is None:
            self._progress_queue = queue

    def task_save_results(self, rgt_obj):
        """"""
        try:
            self._update_progress(ProgressData(percent=25, sender="RGT", message=f"Saving Response Data..."))
            success = rgt_obj.save_results_to_file()
            self._update_progress(ProgressData(percent=95, sender="RGT", message=f"Saving Response Data..."))
            if not success:
                raise ValueError("Error while saving results!")
            task_data = TaskResult(task_id="Save Results", status="Finished", message="Response files successfully saved in 'Output Dir'")
            return True, task_data
        except Exception as err:
            logging.exception("Error: %s", err, extra={'user': 'RGT Logs'})
            return False, ["Download Failed", "Error while saving results!"]

    def task_compute_response(self, rgt_obj):
        """"""
        try:
            rgt_obj.abort = False
            rgt_obj.add_listener(self._update_progress)
            rgt_obj.run_analyzer()
            if rgt_obj.abort:
                raise ValueError("Process aborted")
            rgt_obj.remove_listener(self._update_progress)
            task_data = TaskResult(task_id="Compute Response", status="Finished", message="AC Response computed successfully!", data=rgt_obj)
            return True, task_data
        except ValueError as err:
            logging.exception("Task Aborted: %s", err, extra={'user': 'RGT Logs'})
            # Clean up listeners before exiting
            rgt_obj.remove_listener(self._update_progress)
            return False, ["Computation Aborted", "Error occurred while computing ac-response. Re-upload the CSV files and try again. If error persists then close "
                                                   "the app and try again."]
        except Exception as err:
            logging.exception("Error: %s", err, extra={'user': 'RGT Logs'})
            self._update_progress(ProgressData(type="error", sender="RGT", message=f"Error encountered! Try again."))
            # Clean up listeners before exiting
            rgt_obj.remove_listener(self._update_progress)
            # Emit failure signal (aborted)
            return False, ["Computation Failed", "Error occurred while computing ac-response. Re-upload the CSV files and try again. If error persists then close "
                                                   "the app and try again."]

    def task_upload_csv(self, file_path: str, upload_type: str):
        """"""
        try:
            # 1. Verify if the file exists
            self._update_progress(ProgressData(percent=25, sender="RGT", message=f"Reading File..."))
            success, result = verify_path(file_path)
            if success:
                file_path = result
            else:
                raise ValueError("File Error")

            # 2. Check if the file extension is allowed
            self._update_progress(ProgressData(percent=35, sender="RGT", message=f"Reading File..."))
            file_ext = os.path.splitext(file_path)[1].lower()
            allowed_ext = tuple(ext[1:] if ext.startswith('*.') else ext for ext in ALLOWED_GRAPH_FILE_EXTENSIONS)
            if file_ext not in allowed_ext:
                throw_msg = f"File extension {file_ext} is not allowed. Allowed extensions are {allowed_ext}"
                raise ValueError(throw_msg)

            # 3. Read the file and return graph data
            self._update_progress(ProgressData(percent=50, sender="RGT", message=f"Reading File..."))
            graph_data = csv_to_numpy(file_path)
            self._update_progress(ProgressData(percent=95, sender="RGT", message=f"Reading File..."))
            task_data = TaskResult(task_id="Upload CSV", status="Finished", message="CSV file successfully uploaded!", data=[upload_type, file_path,  graph_data])
            return True, task_data
        except ValueError as err:
            logging.exception("Task Aborted: %s", err, extra={'user': 'RGT Logs'})
            return False, ["File Upload Failed", f"Error while reading file {file_path}!"]
        except Exception as err:
            logging.exception("File Error: %s", err, extra={'user': 'RGT Logs'})
            return False, ["File Error", f"Error reading {file_path}! Try again."]



"""
**ResponseGT**

A software package for computing electrical response properties of graph networks.

"""


# Project Details
__version__ = "1.1.4"
__install_version__ = "1.1.4"
__title__ = "ResponseGT"
__author__ = "William Stephenson and Dickson Owuor"
__credits__ = "The Regents of the University of Michigan"



from PIL import Image, ImageQt  # Import ImageQt for conversion
from PySide6.QtGui import QPixmap
from PySide6.QtQuick import QQuickImageProvider


class ImageProvider(QQuickImageProvider):

    def __init__(self, main_controller):
        super().__init__(QQuickImageProvider.ImageType.Pixmap)
        self._pixmap = QPixmap()
        self._main_ctrl = main_controller
        self._main_ctrl.network_ctrl.changeImageSignal.connect(self.handle_change_image)

    def handle_change_image(self):
        if self._main_ctrl.rgt_obj is None:
            self._main_ctrl.network_ctrl._graph_loaded = False
            self._main_ctrl.network_ctrl._applying_changes = False
            self._main_ctrl.network_ctrl.imageChangedSignal.emit()
            return

        if self._main_ctrl.rgt_obj.network_img is None:
            self._main_ctrl.network_ctrl._graph_loaded = False
            self._main_ctrl.network_ctrl._applying_changes = False
            self._main_ctrl.network_ctrl.imageChangedSignal.emit()
            return

        img_cv = self._main_ctrl.rgt_obj.network_img
        img = Image.fromarray(img_cv)
        self._pixmap = ImageQt.toqpixmap(img)

        # Acknowledge the image load and send the signal to update QML
        self._main_ctrl.network_ctrl._graph_loaded = True
        self._main_ctrl.network_ctrl._applying_changes = False
        self._main_ctrl.network_ctrl.imageChangedSignal.emit()

    def requestPixmap(self, img_id, requested_size, size):
        return self._pixmap



# SPDX-License-Identifier: GNU GPL v3

"""
Entry points that allow users to execute GUI program
"""

import sys
import logging
from .app.gui_app import PySideApp


logger = logging.getLogger("RGT App")
# FORMAT = '%(asctime)s; %(user)s. %(levelname)s: %(message)s'
FORMAT = '%(asctime)s; %(levelname)s: %(message)s'
DATE_FORMAT = '%Y-%m-%d %H:%M:%S'


def main_gui() -> None:
    """
    Start the graphical user interface application.
    :return:
    """
    # Initialize log collection
    logging.basicConfig(level=logging.INFO, format="%(levelname)s: %(message)s", stream=sys.stdout)
    logging.info("RGT application started running...", extra={'user': 'RGT Logs'})

    # Start GUI app
    PySideApp.start()

    # Log to show the App stopped
    logging.info("RGT application stopped running.", extra={'user': 'RGT Logs'})
